"""Provides the :class:`Arc` class."""
from __future__ import annotations

from typing import List, Optional

import numpy as np

from fibomat.linalg import Vector, VectorLike, BoundingBox
from fibomat.shapes.shape import Shape
from fibomat.utils import mod_2pi
from fibomat.shapes.arc_spline import ArcSpline, ArcSplineCompatible


class Arc(Shape, ArcSplineCompatible):  # pylint: disable=too-many-public-methods
    """Circular arc shape.

    Some formulas take from `here <http://www.lee-mac.com/bulgeconversion.html>`_.

    TODO: from_points((-1, 0), (0, 1), (-1, 0)) does not work. Should it!?
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        radius: float,
        start_angle: float, end_angle: float,
        sweep_dir: bool,
        center: Optional[VectorLike] = None,
        description: Optional[str] = None
    ):
        """
        Args:
            radius (float): radius
            start_angle (float): starting angle (measured from pos. x-axis)
            end_angle (float): end angle (measured from pos. x-axis)
            sweep_dir (bool): if True, arc direction is in mathematical positive direction and in math. negative
                              direction if False
            center (VectorLike, optional): center of completed arc, default to (0, 0)
            description (str, optional): description

        .. warning:: `center` is the center of the circle (aka. the completed arc), not centroid!

        """
        super().__init__(description)

        self._r: float = float(radius)
        # make sure that angles are in[0, 2pi]
        self._start_angle: float = mod_2pi(float(start_angle))
        self._end_angle: float = mod_2pi(float(end_angle))
        self._sweep_dir = bool(sweep_dir)
        # self._large_arc = bool(large_arc)
        self._center: Vector = Vector(center) if center is not None else Vector(0, 0)

    @classmethod
    def from_bulge(cls, start: VectorLike, end: VectorLike, bulge: float):
        """
        Construct a curve from start and end points and bulge value.
        See `here <http://www.lee-mac.com/bulgeconversion.html>`_ and `there
        <https://ezdxf.readthedocs.io/en/stable/dxfentities/lwpolyline.html#bulge-value>`_ for details concerning the
        bulge value.

        Args:
            start (VectorLike): start point
            end (VectorLike): end point
            bulge (float): bulge value

        Returns:
            Arc
        """
        start = Vector(start)
        end = Vector(end)
        bulge = float(bulge)

        theta_2 = 2 * np.arctan(bulge)
        radius = (start - end).length / 2 / np.sin(theta_2)
        phi = np.pi / 2 - theta_2 + (end - start).angle_about_x_axis
        # print(theta_2, r, phi)
        center = start+Vector(r=radius, phi=phi)

        # if bulge < 0.:
        #     start_angle = angle_about_xaxis(end - center)
        #     end_angle = angle_about_xaxis(start - center)
        # else:
        start_angle = (start - center).angle_about_x_axis
        end_angle = (end - center).angle_about_x_axis

        return cls(abs(radius), start_angle, end_angle, bool(bulge >= 0), center)

    @classmethod
    def from_points(cls, p1: VectorLike, p2: VectorLike, p3: VectorLike):  # pylint: disable=invalid-name
        """Creates an arc connecting `p1` with `p3` via `p2`.

        Args:
            p1 (VectorLike): start point
            p2 (VectorLike): intermediate point
            p3 (VectorLike): end point

        Returns:
            Arc
        """
        p1 = Vector(p1)
        p2 = Vector(p2)
        p3 = Vector(p3)

        bulge = np.tan((np.pi - (p1 - p2).angle_about_x_axis + (p3 - p2).angle_about_x_axis) / 2)
        # print('bulge', bulge, (np.pi - angle_about_xaxis(p1 - p2) + angle_about_xaxis(p3 - p2)) / 2)

        return cls.from_bulge(p1, p3, bulge)

    @classmethod
    def from_points_center(cls, start: VectorLike, end: VectorLike, center: VectorLike, sweep_dir: bool):
        """Create Arc from start, end, center and sweep_dir

        Args:
            start (VectorLike): start point
            end (VectorLike): end point
            center (VectorLike): center
            sweep_dir (bool): sweep_dir

        Returns:
            Arc
        """
        start = Vector(start)
        end = Vector(end)
        center = Vector(center)

        start_angle = (start - center).angle_about_x_axis
        end_angle = (end - center).angle_about_x_axis

        radius = np.linalg.norm(start - center)

        return cls(radius=radius, start_angle=start_angle, end_angle=end_angle, sweep_dir=sweep_dir, center=center)

    @classmethod
    def from_points_center_tangent(
        cls,
        start: VectorLike,
        end: VectorLike,
        center: VectorLike,
        *, unit_tangent_start=None, unit_tangent_end=None
    ):
        """Create Arc from start, end, center and tangent at start or end. The sweep_dir is calculated automatically.

        Args:
            start (VectorLike): start point
            end (VectorLike): end point
            center (VectorLike): center
            unit_tangent_start (VectorLike, optional): unit tangent at start
            unit_tangent_end (VectorLike, optional): unit tangent at end

        Returns:
            Arc

        Raises:
            ValueError: Raised of none or both of unit_tangent_start and unit_tangent_end are defined.
        """
        start = Vector(start)
        end = Vector(end)
        center = Vector(center)

        start_angle = (start - center).angle_about_x_axis
        end_angle = (end - center).angle_about_x_axis

        radius = np.linalg.norm(start - center)

        if unit_tangent_start is not None and unit_tangent_end is None:
            unit_tangent_start = Vector(unit_tangent_start)

            sweep_matrix = np.ones(shape=(3, 3), dtype=float)
            sweep_matrix[0, 1:] = np.asarray(start)
            sweep_matrix[1, 1:] = np.asarray(start + unit_tangent_start)
            sweep_matrix[2, 1:] = np.asarray(end)
        elif unit_tangent_start is None and unit_tangent_end is not None:
            unit_tangent_end = Vector(unit_tangent_end)

            sweep_matrix = np.ones(shape=(3, 3), dtype=float)
            sweep_matrix[0, 1:] = np.asarray(start)
            sweep_matrix[1, 1:] = np.asarray(end)
            sweep_matrix[2, 1:] = np.asarray(end + unit_tangent_end)
        elif unit_tangent_start is None and unit_tangent_end is None:
            raise ValueError('Anyone of unit_tangent_start or unit_tangent_end must be defined.')
        else:
            raise ValueError('unit_tangent_start and unit_tangent_end cannot be defined both.')

        sweep_dir = bool(np.linalg.det(sweep_matrix) > 0)

        arc = cls(radius=radius, start_angle=start_angle, end_angle=end_angle, sweep_dir=sweep_dir, center=center)

        if unit_tangent_start is not None:
            if not np.allclose(arc.unit_tangent_start, unit_tangent_start):
                raise RuntimeError
        else:
            if not np.allclose(arc.unit_tangent_end, unit_tangent_end):
                raise RuntimeError

        return arc

    def to_arc_spline(self) -> ArcSpline:
        is_closed = self.is_closed

        if abs(self.bulge) > 1:
            arcs = self.split()
            vertices = [(*arcs[0].start, arcs[0].bulge), (*arcs[1].start, arcs[1].bulge)]
            if not is_closed:
                vertices.append((*arcs[1].end, 0.))
        else:
            vertices = [(*self.start, self.bulge), (*self.end, 0.)]

        return ArcSpline(vertices, is_closed, self.description)

    @property
    def start(self) -> Vector:
        """Start point of arc

        Access:
            get

        Returns:
            Vector
        """
        return Vector(r=self._r, phi=self._start_angle)+self._center

    @property
    def end(self) -> Vector:
        """End point of arc

         Access:
             get

         Returns:
             Vector
         """
        return Vector(r=self._r, phi=self._end_angle)+self._center

    @property
    def start_angle(self) -> float:
        """Start angle of arc (measured from pos. x. axis).

        Access:
            get

        Returns:
            float
        """
        return self._start_angle

    @property
    def end_angle(self):
        """End angle of arc (measured from pos. x. axis).

        Access:
            get

        Returns:
            float
        """
        return self._end_angle

    @property
    def sweep_dir(self) -> bool:
        """If `True`, arc direction is in mathematical positive direction and in math. negative direction if `False`

        Access:
            get

        Returns:
            bool
        """
        return self._sweep_dir

    @property
    def radius(self) -> float:
        """Arc radius

        Access:
            get

        Returns:
            float
        """
        return self._r

    @property
    def theta(self) -> float:
        """Absolute value of  enclosed angle

        Access:
            get

        Returns:
            float
        """
        angle = np.abs(self._end_angle - self._start_angle)
        if (self._sweep_dir and self._start_angle >= self._end_angle) \
           or (not self._sweep_dir and self._end_angle >= self._start_angle):
            angle = 2*np.pi - angle

        return angle

    @property
    def bulge(self) -> float:
        """Bulge value = tan(theta/4). bulge > 0 arc goes in math. positve direction and for bulge < 0 in math.
        neg. direction

        Access:
            get

        Returns:
            float
        """
        if self._sweep_dir:
            sign = 1
        else:
            sign = -1
        return sign * np.tan(self.theta / 4.)

    @property
    def center(self) -> Vector:
        """Center of enclosing circle.

        Access:
            get

        Returns:
            Vector
        """
        return self._center

    @property
    def midpoint(self) -> Vector:
        """Midpoint of arc.

        Access:
            get

        Returns:
            Vector
        """
        sign = 1. if self._sweep_dir else -1.
        return self.start.rotated(sign * self.theta/2, origin=self._center)

    @property
    def unit_tangent_start(self) -> Vector:
        """Unit tangent at start.

        Access:
            get

        Returns:
            Vector
        """
        sign = 1. if self._sweep_dir else -1.
        return sign * Vector(-np.sin(self._start_angle), np.cos(self._start_angle))

    @property
    def unit_tangent_end(self) -> Vector:
        """Unit tangent at end.

        Access:
            get

        Returns:
            Vector
        """
        sign = 1. if self._sweep_dir else -1.
        return sign * Vector(-np.sin(self._end_angle), np.cos(self._end_angle))

    def unit_tangent_at(self, angle):
        """Unit tangent at specific angle.

        Args:
            angle (float): 0 < angle < self.theta

        Returns:
            Vector
        """
        if angle > self.theta:
            raise ValueError('angle > self.theta')

        sign = 1. if self._sweep_dir else -1.
        phi = self._start_angle + sign * angle

        return sign * Vector(-np.sin(phi), np.cos(phi))

    @property
    def length(self):
        """Length of arc.

        Access:
            get

        Returns:
            float
        """
        return self.radius * self.theta

    def split(self) -> List[Arc]:
        """Return two arcs if theta > np.pi (hence the abs(bulge) value of new arcs will be smaller than 1).

        Returns:
            List[Arc]: if theta > np.pi list contains two elements and one otherwise.
        """
        # pylint: disable=protected-access
        theta = self.theta
        if theta > np.pi:
            arc_1 = self.clone()
            arc_2 = self.clone()

            if self._sweep_dir:
                arc_1._end_angle = mod_2pi(arc_1._start_angle + theta / 2)
                arc_2._start_angle = arc_1._end_angle
            else:
                arc_1._end_angle = mod_2pi(arc_1._start_angle - theta / 2)
                arc_2._start_angle = arc_1._end_angle
            return [arc_1, arc_2]

        return [self.clone()]

    def split_at(self, angle: float) -> List[Arc]:
        """Split the arc in two halves ([0, angle], [angle, theta]).

        Args:
            angle (float): split angle.

        Returns:
            List[Arc]

        Raises:
            ValueError: Raised if angle < 0 or angle > self.theta
        """
        # pylint: disable=protected-access

        if angle < 0 or angle > self.theta:
            raise ValueError('angle < 0 or angle > self.theta')

        arc_1 = self.clone()
        arc_2 = self.clone()

        if self._sweep_dir:
            arc_1._end_angle = mod_2pi(arc_1._start_angle + angle)
            arc_2._start_angle = arc_1._end_angle
        else:
            arc_1._end_angle = mod_2pi(arc_1._start_angle - angle)
            arc_2._start_angle = arc_1._end_angle
        return [arc_1, arc_2]

    # def reverse(self) -> None:
    #     """Reverse the arc in-place.
    #
    #     Returns:
    #         None
    #     """
    #     self._start_angle, self._end_angle = self._end_angle, self._start_angle
    #     self._sweep_dir = not self._sweep_dir

    def reversed(self) -> Arc:
        """Return a reversed version of the arc

        Returns:
            Arc
        """
        # pylint: disable=protected-access
        copy = self.clone()
        copy._start_angle, copy._end_angle = copy._end_angle, copy._start_angle
        copy._sweep_dir = not copy._sweep_dir
        return copy

    # def clone(self) -> Arc:
    #     return self.__class__(self._r, self._start_angle, self._end_angle, self._sweep_dir, self._center.clone())
    #
    # __copy__ = clone

    def __repr__(self) -> str:
        return 'Arc(r={!r}, start_angle={!r}, end_angle={!r}, sweep_dir={!r}, center={!r})'.format(
            self.radius, self.start_angle, self.end_angle, self.sweep_dir, self.center
        )

    @property
    def bounding_box(self) -> BoundingBox:
        # https://stackoverflow.com/questions/1336663/2d-bounding-box-of-a-sector
        def axis_intersection(angle: float):
            return self._center+Vector(self._r, 0).rotated(angle)

        if np.isclose(self.theta, 2*np.pi):
            r_vec = Vector(self._r, self._r)
            return BoundingBox(self.center-r_vec, self.center+r_vec)

        points = [self._center, self.start, self.end]

        for axis_angle in [0., np.pi/2, np.pi, 3/2 * np.pi, 2*np.pi]:
            if self._sweep_dir:
                if self._end_angle >= self._start_angle:
                    if self._end_angle >= axis_angle >= self._start_angle:
                        points.append(axis_intersection(axis_angle))
                else:  # self._end_angle < self._start_angle
                    if self._end_angle >= axis_angle or self._start_angle <= axis_angle:
                        points.append(axis_intersection(axis_angle))
            else:
                if self._start_angle >= self._end_angle:
                    if self._start_angle >= axis_angle >= self._end_angle:
                        points.append(axis_intersection(axis_angle))
                else:  # self._start_angle < self._end_angle
                    if self._end_angle <= axis_angle or self._start_angle >= axis_angle:
                        points.append(axis_intersection(axis_angle))

        return BoundingBox.from_points(points)

    @property
    def is_closed(self) -> bool:
        return np.isclose(self.theta, 2*np.pi)

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._center += Vector(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        theta = float(theta)

        self._start_angle = mod_2pi(self._start_angle + theta)
        self._end_angle = mod_2pi(self._end_angle + theta)

    def _impl_scale(self, fac: float) -> None:
        self._r *= float(fac)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        # pylint: disable=protected-access

        mirrored_arc = self.__class__.from_points_center(
            self.start.mirrored(mirror_axis),
            self.end.mirrored(mirror_axis),
            self._center.mirrored(mirror_axis),
            not self._sweep_dir
        )
        self._start_angle = mirrored_arc._start_angle
        self._end_angle = mirrored_arc._end_angle
        self._sweep_dir = mirrored_arc._sweep_dir
        self._center = mirrored_arc._center



from typing import Optional

import numpy as np

import pyhershey

from fibomat.shapes.polyline import Polyline
from fibomat.shapes.polygon import Polygon
from fibomat.layout import Group, DimGroup
from fibomat.linalg import Vector, translate
from fibomat.linalg.helpers import make_perp_vector, GeomLine
from fibomat.units import U_
# from fibomat.curve_tools import


class Text(Group):
    """"""

    @staticmethod
    def _add_stroke(polyline: Polyline, stroke_width: float) -> Polygon:
        def make_normals(p1: Vector, p2: Vector):
            normal = make_perp_vector(p2-p1).normalized_to(stroke_width)
            return normal, -1 * normal

        def make_offset_segments(p1: Vector, p2: Vector):
            normals = make_normals(p1, p2)

            length = np.linalg.norm(p1 - p2)

            return (
                GeomLine(Vector((p2 + normals[0]) - (p1 + normals[0])).normalized_to(length), (p1 + normals[0])),
                GeomLine(Vector((p2 + normals[1]) - (p1 + normals[1])).normalized_to(length), (p1 + normals[1]))
            )

        def make_points(segments_, prev_segments_):
            if segments_[0].parallel_to(prev_segments_[0]):
                return [segments_[0](0)], [segments_[1](0)]
            else:
                param = segments_[0].intersect_at_param(prev_segments_[0])

                if 0 < param < 1:
                    return [segments_[0](param)], [prev_segments_[1](1), segments_[1](0)]
                else:
                    param = segments_[1].intersect_at_param(prev_segments_[1])

                    return [prev_segments_[0](1), segments_[0](0)], [segments_[1](param)]

        polyline_points = polyline.points

        if len(polyline_points) < 2:
            raise RuntimeError

        closed = np.allclose(polyline_points[0], polyline_points[-1])

        points_first_side = []
        points_second_side = []

        first_segments = make_offset_segments(polyline_points[0], polyline_points[1])

        if not closed:
            points_first_side.append(first_segments[0](0))
            points_second_side.append(first_segments[1](0))

        prev_segments = first_segments

        for i in range(1, len(polyline_points) - 1):
            segments = make_offset_segments(polyline_points[i], polyline_points[i+1])

            first, second = make_points(segments, prev_segments)

            points_first_side.extend(first)
            points_second_side.extend(second)

            prev_segments = segments

        if closed:
            first, second = make_points(first_segments, prev_segments)
            points_first_side.extend(first)
            points_second_side.extend(second)

            points_first_side.insert(0, first[-1])
            points_second_side.insert(0, second[-1])
        else:
            points_first_side.append(prev_segments[0](1))
            points_second_side.append(prev_segments[1](1))

        return Polygon(points_first_side + list(reversed(points_second_side)))

    def __init__(
        self,
        text: str,
        font_size: float = 1,
        stroke_width: Optional[float] = None,
        text_alignment: Optional[str] = None,
        description: Optional[str] = None,
        mapping: Optional[str] = None
    ):
        """
        Args:
            text (str): text
            font_size (float, optional):
                font size (directly correspond to the height of upper case letters). Default to 1.
            stroke_width (float, optional):
                the stroke width of the glyph segments. If None, the glyph segments are given by 1d polylines.
                Default to None.
            text_alignment (str, optional): the text alignment. Can be "left", "center", "right". Default to "left".
        """

        mapping = mapping or 'roman_simplex'

        font_size /= 21
        advance_height = 1.6 * 21 * font_size

        if not text_alignment:
            text_alignment = 'left'

        shaped_text_lines = [
            pyhershey.shape_text(
                text_line, advance_height=0, mapping=mapping, font_size=font_size, text_align=text_alignment
            ) for text_line in text.split('\n')
        ]

        glyph_groups = []

        y_shift = 0
        anchors = []

        for shaped_text_line in shaped_text_lines:
            for shaped_glyph in shaped_text_line:
                glyph_polylines = []

                for segment in shaped_glyph['glyph'].segments:

                    # if np.allclose(segment[0], segment[-1]):
                    #     seg_shape = Polygon(segment)
                    # else:
                    #     seg_shape = Polyline(segment)

                    if stroke_width:
                        glyph_polylines.append(self._add_stroke(Polyline(segment), stroke_width))
                    else:
                        glyph_polylines.append(Polyline(segment))

                if glyph_polylines:
                    glyph_groups.append(
                        Group(glyph_polylines).transformed(translate(shaped_glyph['pos']) | translate((0, y_shift)))
                    )

            anchor_left = Vector(shaped_text_line[0]['pos']) + (0, y_shift)
            anchor_right = Vector(shaped_text_line[-1]['pos']) + (shaped_text_line[-1]['glyph'].advance_width, y_shift)
            anchor_center = Vector((anchor_left.x + anchor_right.x) / 2, y_shift)

            y_shift -= advance_height

            anchors.append(
                {'left': anchor_left, 'center': anchor_center, 'right': anchor_right}
            )

        self._n_lines = len(shaped_text_lines)
        self._anchors = anchors

        # glyph_group = Group(glyph_groups)

        super().__init__(glyph_groups, description)

    @property
    def n_lines(self) -> int:
        """int: number of text lines"""
        return self._n_lines

    def baseline_anchor(self, pos: str, i_line: int = 0) -> Vector:
        """Return position of certain base line points.

        Args:
            pos (str): position on the base line. Can be "left", "center" or "right".
            i_line (int, optional):
                index of the base line to be used. Can be indexed like any array (0 is the first element, -1 the last
                etc.). Default to 0.

        Returns:

        """
        return self._anchors[i_line][pos]

    def __mul__(self, other):
        if isinstance(other, U_):
            # from fibomat.layout.groups.dim_group import DimGroup
            return DimText(self.elements, self._anchors, other, description=self.description)
        raise NotImplementedError


class DimText(DimGroup):
    def __init__(self, elements, anchors, unit, description):
        super().__init__([elem * unit for elem in elements], description)

        self._anchors = anchors
        self._unit = unit

    @property
    def n_lines(self) -> int:
        return len(self._anchors)

    def baseline_anchor(self, pos: str, i_line: int = 0) -> Vector:
        return self._anchors[i_line][pos] * self._unit


# def shape_text(
#
# ) -> Group:
#     """Shape a given text.
#     This create a group containing glyphs where each glyph is given as a subgroup.
#     ``text`` can be any string but may only contain printable ASCII characters and "°". New lines can be introduced
#     with "\\n".
#
#     The pivot of the returned group is set to start of the baseline of the first text row. If ``text_align`` is "center"
#     or "right", the pivot is at the center or right side of the baseline, respectively.
#
#     Args:
#         text (str): text
#         font_size (float, optional):
#             font size (directly correspond to the height of upper case letters). Default to 1.
#         stroke_width (float, optional):
#             the stroke width of the glyph segments. If None, the glyph segments are given by 1d polylines.
#             Default to None.
#         text_alignment (str, optional): the text alignment. Can be "left", "center", "right". Default to "left".
#
#     Returns:
#         Group: grouped shaped text as
#     """
#
#     font_size /= 21
#     advance_height = 1.5 * 21 * font_size
#
#     if not text_alignment:
#         text_alignment = 'left'
#
#     shaped_glyphs = pyhershey.shape_text(
#         text, advance_height=advance_height, mapping='roman_simplex', font_size=font_size, text_align=text_alignment
#     )
#
#     glyph_groups = []
#
#     for shaped_glyph in shaped_glyphs:
#         glyph_polylines = []
#
#         for segment in shaped_glyph['glyph'].segments:
#             if stroke_width:
#                 glyph_polylines.append(_add_stroke(Polyline(segment), stroke_width))
#             else:
#                 glyph_polylines.append(Polyline(segment))
#
#         if glyph_polylines:
#             glyph_groups.append(Group(glyph_polylines).translated(shaped_glyph['pos']))
#
#     glyph_group = Group(glyph_groups)
#
#     glyph_group.pivot =
#
#     return glyph_group.translated_to((0, 0))



from typing import Iterable, Union, Callable, Optional, Tuple, List

from fibomat.shapes import HollowArcSpline, ArcSplineCompatible, Rect
from fibomat.linalg import Transformable, Vector, VectorLike
from fibomat.layout.lattices.lattice_base import LatticeBaseMixin
from fibomat.layout.groups.group import Group
from fibomat.units import U_


class Lattice(Group, LatticeBaseMixin):
    def __init__(self, elements: List[Transformable], description: Optional[str] = None):
        super().__init__(elements, description)

    @classmethod
    def generate_rect(
        cls,
        nu: int, nv: int,
        du: float, dv: float,
        element: Union[Transformable, Callable],
        center: Optional[VectorLike] = None,
        predicate: Optional[Union[Callable, List[Callable]]] = None,
        explode: bool = False,
        remove_outliers: bool = False,
    ):
        nu = int(nu)
        nv = int(nv)

        if nu < 1 or nv < 1:
            raise ValueError('nu and nv must be at least 1.')

        du = float(du)
        dv = float(dv)

        seed_u = du / 2 if nu % 2 == 0 else 0
        seed_v = dv / 2 if nv % 2 == 0 else 0

        seed = Vector(seed_u, seed_v)

        return cls.generate(
            Rect(width=du*nu, height=dv*nv), (du, 0), (0, -dv), element, center, predicate, explode, remove_outliers, seed
        )

    @classmethod
    def generate(
        cls,
        boundary: Union[HollowArcSpline, ArcSplineCompatible],
        u: VectorLike, v: VectorLike,
        element: Union[Transformable, Callable[[Tuple[float, float], Tuple[int, int]], Optional[Transformable]]],
        center: Optional[VectorLike] = None,
        predicate: Optional[Union[Callable, List[Callable]]] = None,
        explode: bool = False,
        remove_outliers: bool = False,
        seed: Optional[VectorLike] = None
        # break_layouts: bool = False
    ):
        u = Vector(u)
        v = Vector(v)

        center = Vector(center)

        if callable(element):
            element_gen = element
        else:
            def element_gen(*args):
                return element

        if seed is None:
            seed = center
        else:
            seed = Vector(seed)

        elements = cls._generate_impl(
            boundary, u, v, center, element_gen, predicate, explode, remove_outliers, lambda x: x, lambda x: x, seed
        )

        return cls(elements)

    def __mul__(self, other):
        if isinstance(other, U_):
            from fibomat.layout.lattices.dim_lattice import DimLattice
            return DimLattice([elem * other for elem in self.elements], description=self.description)
        raise NotImplementedError




"""
This submodule implements two exemplary backends.
The first one, the bokeh backend, can be used to visualize a project. It uses the bokeh library to create a plot which
can be viewed with a web browser.
The second one, the spotlist backend, rasterizes all shapes and creates a list of dwell points and times.
"""


from fibomat.backend import registry

from fibomat.default_backends.bokeh_backend import BokehBackend, StubRasterStyle
# from fibomat.default_backends.donothing_backend import DoNothingBackend
from fibomat.default_backends.spotlist_backend import SpotListBackend
from fibomat.default_backends.svg_backend import SVGBackend
from fibomat.default_backends.bitmap_backend import BitmapBackend


registry.register(BokehBackend, BokehBackend.name)
registry.register(SpotListBackend, SpotListBackend.name)
# registry.register(DoNothingBackend, DoNothingBackend.name)

__all__ = ['BokehBackend', 'StubRasterStyle', 'SpotListBackend']



"""Provide the :func:`combine_curves` function."""
from typing import Dict, List

from fibomat.shapes import ArcSpline
from fibomat import _libfibomat


def combine_curves(curve_1: ArcSpline, curve_2: ArcSpline, mode: str) -> Dict[str, List[ArcSpline]]:
    """Combine two curves. The combining method is given by `mode`.

    `mode` can be:

        - `union`
        - `xor`
        - `exclude`
        - `intersect`

    See Usage guide for further details.

    Args:
        curve_1 (ArcSpline): first curve
        curve_2 (ArcSpline): second curve
        mode (str): combining mode

    Returns:
        List[ArcSpline]: list of new ArcSpline
    """
    remaining, subtracted = _libfibomat.combine_curves(curve_1.arc_spline_impl, curve_2.arc_spline_impl, mode)

    return {
        'remaining': [ArcSpline(raw) for raw in remaining],
        'subtracted': [ArcSpline(raw) for raw in subtracted]
    }


# def trim_curves(curve_1: shapes.Curve, curve_2: shapes.Curve) -> Dict[str, List[shapes.Curve]]:
#     """Trim curve `curve_2` on curve `curve_1`.
#
#     .. warning:: no working currently
#
#     Args:
#         curve_1 (shapes.Curve): trimming geometry, curve must be closed.
#         curve_2 (shapes.Curve): curve to be trimmed
#
#     Returns:
#         Dict[str, List[shapes.Curve]]: dict with keys `inner` and `outer`. Curves in former category lie in the inside
#                                        of `curve_1` and curves in latter category on the outside.
#     """
#
#     trimmed_curves = _libfibomat.trim_curves(curve_1._curve, curve_2._curve)
#     return {
#         'inside': [shapes.Curve.from_raw(raw) for raw in trimmed_curves[0]],
#         'outside': [shapes.Curve.from_raw(raw) for raw in trimmed_curves[1]]
#     }



from __future__ import annotations
from typing import Tuple, List, TYPE_CHECKING

import enum

import numpy as np

if TYPE_CHECKING:
    from fibomat.shapes.parametric_curve import ParametricCurve

from .utils import _make_perp_vector


@enum.unique
class _CircleRelations(enum.Enum):
    CIRCLE_1_IN_CIRCLE_2 = enum.auto()
    CIRCLE_2_IN_CIRCLE_1 = enum.auto()
    IDENTICAL = enum.auto()
    NO_INTERSECTION = enum.auto()
    OVERLAP = enum.auto()


@enum.unique
class _CurvatureState(enum.Enum):
    INCREASING = enum.auto()
    DECREASING = enum.auto()
    CONSTANT = enum.auto()
    INFINITE = enum.auto()

    UNDEFINED = enum.auto()

    # at begin of new segment
    NONE = enum.auto()


@enum.unique
class _FlushMode(enum.Enum):
    NO_FLUSH = enum.auto()
    FLUSH_WITH_TRANSITION = enum.auto()
    FLUSH_WITHOUT_TRANSITION = enum.auto()


@enum.unique
class IntervalType(enum.Enum):
    LINE = enum.auto()
    ARC = enum.auto()
    BIARC = enum.auto()
    SPIRAL = enum.auto()


def _osculating_circle(param_curve: 'ParametricCurve', parameter: float):
    curvature = param_curve.curvature(parameter)
    radius = abs(1 / curvature)

    tangent = param_curve.df(parameter)
    tangent /= np.linalg.norm(tangent)

    t_perp = np.sign(curvature) * _make_perp_vector(tangent)
    center_osc_circle = param_curve.f(parameter) + radius * t_perp

    return center_osc_circle, radius


def _circle_relation(circle_1: Tuple[np.ndarray, float], circle_2: Tuple[np.ndarray, float]):
    c_1, r_1 = circle_1
    c_2, r_2 = circle_2

    d = np.linalg.norm(c_1 - c_2)

    if np.allclose(c_1, c_2) and np.allclose(r_1, r_2):
        return _CircleRelations.IDENTICAL
    if r_1 >= d + r_2:
        return _CircleRelations.CIRCLE_2_IN_CIRCLE_1
    elif r_2 >= d + r_1:
        return _CircleRelations.CIRCLE_1_IN_CIRCLE_2
    elif d >= r_1 + r_2 + d:
        return _CircleRelations.NO_INTERSECTION
    else:
        return _CircleRelations.OVERLAP


def interval_type(n_points: int, curvature_state: _CurvatureState):
    if n_points < 2:
        raise RuntimeError
    elif n_points == 2:
        return IntervalType.BIARC
    else:
        if curvature_state == _CurvatureState.INFINITE:
            return IntervalType.LINE
        elif curvature_state == _CurvatureState.CONSTANT:
            return IntervalType.ARC
        elif curvature_state in (_CurvatureState.INCREASING, _CurvatureState.DECREASING):
            return IntervalType.SPIRAL
        else:
            raise RuntimeError


def intervals_by_osculating_circles(
    param_curve: ParametricCurve, parameter_values: np.ndarray, epsilon: float
) -> List[Tuple[IntervalType, np.ndarray]]:

    # TODO: INFINITE should be really ZERO_CURVATURE

    if len(parameter_values) == 0:
        raise RuntimeError

    prev_osculating_circle = None
    curvature_state = _CurvatureState.NONE

    intervals = []
    current_interval = []

    for i, param in enumerate(parameter_values):
        if np.allclose(param_curve.df(param), 0):
            if i == len(parameter_values) - 1:
                k = param_curve.curvature(param - epsilon)
            else:
                k = param_curve.curvature(param + epsilon)
        else:
            k = param_curve.curvature(param)

        flush = _FlushMode.NO_FLUSH

        if np.isclose(k, 0.):
            if curvature_state == _CurvatureState.NONE:
                current_interval.append(param)
                curvature_state = _CurvatureState.UNDEFINED
            elif curvature_state == _CurvatureState.UNDEFINED:
                current_interval.append(param)
                curvature_state = _CurvatureState.INFINITE
            elif curvature_state == _CurvatureState.INFINITE:
                current_interval.append(param)
            elif curvature_state == _CurvatureState.DECREASING:
                # finish current interval and start a new one but without a connecting biarc segment
                flush = _FlushMode.FLUSH_WITHOUT_TRANSITION
            else:
                flush = _FlushMode.FLUSH_WITH_TRANSITION
        else:
            osculating_circle = _osculating_circle(param_curve, param)

            if curvature_state == _CurvatureState.NONE:
                current_interval.append(param)
                curvature_state = _CurvatureState.UNDEFINED
            elif curvature_state == _CurvatureState.INFINITE:
                # this could be handled better most likely
                flush = _FlushMode.FLUSH_WITH_TRANSITION
            elif curvature_state == _CurvatureState.CONSTANT:
                relation = _circle_relation(osculating_circle,  prev_osculating_circle)

                if relation == _CircleRelations.IDENTICAL:
                    current_interval.append(param)
                else:
                    flush = _FlushMode.FLUSH_WITH_TRANSITION
            elif curvature_state == _CurvatureState.UNDEFINED and prev_osculating_circle is None:
                # Can this happen at all?
                flush = _FlushMode.FLUSH_WITHOUT_TRANSITION
            elif curvature_state == _CurvatureState.UNDEFINED:
                current_interval.append(param)

                # determine curvature state
                relation = _circle_relation(osculating_circle,  prev_osculating_circle)
                if relation == _CircleRelations.NO_INTERSECTION:
                    raise RuntimeError(
                        'No intersection between osculating circles. Try to decrease the rasterization pitch.'
                    )
                elif relation == _CircleRelations.IDENTICAL:
                    current_interval.append(param)
                    curvature_state = _CurvatureState.CONSTANT
                    # if np.isclose(osculating_circle[1], prev_osculating_circle[1]):
                    #     curvature_state = _CurvatureState.CONSTANT
                    # else:
                    #     flush = _FlushMode.FLUSH_WITHOUT_TRANSITION
                else:
                    if relation == _CircleRelations.CIRCLE_1_IN_CIRCLE_2:
                        curvature_state = _CurvatureState.INCREASING
                    else:
                        curvature_state = _CurvatureState.DECREASING
            elif curvature_state in (_CurvatureState.INCREASING, _CurvatureState.DECREASING):
                relation = _circle_relation(osculating_circle,  prev_osculating_circle)

                if relation == _CircleRelations.NO_INTERSECTION:
                    raise RuntimeError(
                        'No intersection between osculating circles. Try to decrease the rasterization pitch.'
                    )
                elif curvature_state == _CurvatureState.INCREASING and relation == _CircleRelations.CIRCLE_1_IN_CIRCLE_2:
                    current_interval.append(param)
                    prev_osculating_circle = osculating_circle
                elif curvature_state == _CurvatureState.DECREASING and relation == _CircleRelations.CIRCLE_2_IN_CIRCLE_1:
                    current_interval.append(param)
                    prev_osculating_circle = osculating_circle
                else:  # _CircleRelations.OVERLAP or something else => make a transition segment and new interval
                    flush = _FlushMode.FLUSH_WITH_TRANSITION
            else:
                raise RuntimeError('Failed.')

            prev_osculating_circle = osculating_circle

        if flush == _FlushMode.FLUSH_WITH_TRANSITION:
            prev_osculating_circle = None

            intervals.append((interval_type(len(current_interval), curvature_state), np.array(current_interval)))
            intervals.append((IntervalType.BIARC, np.array([parameter_values[i-1], param])))

            curvature_state = _CurvatureState.NONE
            prev_osculating_circle = None

            current_interval = [param]
        elif flush == _FlushMode.FLUSH_WITHOUT_TRANSITION:
            prev_osculating_circle = None

            current_interval.append(param)
            intervals.append((interval_type(len(current_interval), curvature_state), np.array(current_interval)))

            curvature_state = _CurvatureState.NONE
            prev_osculating_circle = None

            current_interval = [param]

    if flush == _FlushMode.NO_FLUSH:
        intervals.append((interval_type(len(current_interval), curvature_state), np.array(current_interval)))

    return intervals



"""Provide the :class:`RasterStyle` class."""
from __future__ import annotations

import abc
import copy

from fibomat.rasterizedpattern import RasterizedPattern
from fibomat.shapes.dim_shape import DimShape
from fibomat.units import LengthUnit, TimeUnit
from fibomat.mill import Mill


class RasterStyle(abc.ABC):
    """Base class to define a raster style.

    All raster styles have the required parameters as attributes (e.g. pitch, ...).
    Further, all raster styles must implement a rasterize method which applies the raster style to a shape.
    """
    def __repr__(self) -> str:
        return self.__class__.__name__

    def clone(self) -> RasterStyle:
        """Deep copy the raster style.

        Returns:
            RasterStyle
        """
        return copy.deepcopy(self)

    @property
    @abc.abstractmethod
    def dimension(self) -> int:
        """Returns the required dimensionality a shape must have to apply this raster style
        Returns:

        """
        raise NotImplementedError

    @abc.abstractmethod
    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:
        """Rasterize the given shape.

        The number of repeats and dwell_time defined in the mill object must be applied!
        The points and dwell times in the returned RasterizedPoints object must be scaled according to out_length_unit
        and out_time_unit.

        Args:
            dim_shape (DimObjLike[Shape, LengthUnit]): shpae with length unit to be rasterized.
            mill (Mill): mill
            out_length_unit (LengthUnit): length unit of returned RasterizedPoints
            out_time_unit (TimeUnit): time unit of returned RasterizedPoints

        Returns:
            RasterizedPoints
        """
        raise NotImplementedError

    # @abc.abstractmethod
    # def explode(
    #     self,
    #     dim_shape: DimObjLike[Shape, LengthUnit],
    #     mill: Mill,
    #     out_length_unit: LengthUnit,
    #     out_time_unit: TimeUnit
    # ) -> Sequence[Tuple[Shape, RasterStyle]]:
    #     """
    #     Apply raster style to given shape and break it down to "simpler", lower dimensional operation.
    #     E.g. a Rect shape could be exploded into a lot of individual lines.
    #
    #     If no reduction is possible, the same shape (maybe scaled, according to out_length_unit) and RasterStyle will
    #     be returned.
    #
    #     Args:
    #         dim_shape:
    #         mill:
    #         out_length_unit:
    #         out_time_unit:
    #
    #     Returns:
    #
    #     """
    #     raise NotImplementedError



"""Provide LineByLine raster style."""

import numpy as np

from fibomat.rasterizedpattern import RasterizedPattern
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, LengthQuantity, has_length_dim, scale_to, scale_factor
from fibomat.mill import Mill
from fibomat.shapes import Shape, RasterizedPoints, DimShape
from fibomat.raster_styles.scansequence import ScanSequence, _make_scan_index_sequence, _apply_scan_sequence
from fibomat.raster_styles.one_d import Curve
from fibomat.curve_tools import fill_with_lines, rasterize
from fibomat.shapes._line_non_continuous import LineNonContinuous


class LineByLine(RasterStyle):
    def __init__(
        self,
        line_pitch: LengthQuantity,
        scan_sequence: ScanSequence,
        alpha: float,
        invert: bool,
        line_style: RasterStyle
    ):
        if not has_length_dim(line_pitch):
            raise ValueError('line_pitch must have dimension [length].')
        self._line_pitch = line_pitch

        self._alpha = alpha
        self._invert = invert

        self._scan_sequence = scan_sequence

        if line_style.dimension != 1:
            raise ValueError('line_style must have dimension == 1')

        # if not line_style.scan_sequence == ScanSequence.CONSECUTIVE:
        #     raise NotImplementedError

        self._line_style = line_style

    @property
    def dimension(self) -> int:
        return 2

    @property
    def line_pitch(self):
        return self._line_pitch

    @property
    def scan_sequence(self):
        return self._scan_sequence

    @property
    def alpha(self):
        return self._alpha

    @property
    def invert(self):
        return self._invert

    @property
    def line_style(self):
        return self._line_style

    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:
        # dim_shape = DimObj.create(dim_shape)

        filling_rows = fill_with_lines(
            dim_shape.shape.to_arc_spline(), scale_to(dim_shape.unit, self._line_pitch), self._alpha, self._invert
        )

        return _apply_scan_sequence(
            filling_rows=[LineNonContinuous(row) for row in filling_rows],
            filling_rows_length_unit=dim_shape.unit,
            scan_sequence=self._scan_sequence,
            line_style=self._line_style,
            mill=mill,
            out_length_unit=out_length_unit,
            out_time_unit=out_time_unit
        )

        # rasterized_lines = []
        #
        # if self._scan_sequence == ScanSequence.CROSSECTION:
        #     line_mill = mill
        # else:
        #     line_mill = Mill(mill.dwell_time, repeats=1)
        #
        # #     line_mill = mill
        # # else:
        # # line_mill = Mill(mill.dwell_time, repeats=1)
        #
        # for row in filling_rows:
        #     # rasterized_row = []
        #
        #     row = LineNonContinuous(row)
        #
        #     #for line in row:
        #     rasterized_lines.append(
        #         self._line_style.rasterize(
        #             (row, dim_shape.unit),
        #             mill=line_mill,
        #             out_length_unit=out_length_unit,
        #             out_time_unit=out_time_unit
        #         ).dwell_points
        #     )
        #
        #     # rasterized_lines.append(rasterized_row))
        #
        # scan_lines = []
        #
        # for index, direction in zip(*_make_scan_index_sequence(len(rasterized_lines), mill.repeats, self._scan_sequence)):
        #     if not direction:
        #         scan_lines.append(rasterized_lines[index][::-1])
        #     else:
        #         scan_lines.append(rasterized_lines[index])
        #
        # # points = np.concatenate(scan_lines)
        #
        # # points._dwell_points[:, :2] *= scale_factor(out_length_unit, dim_shape.unit)
        # # points._dwell_points[:, 2] = scale_to(out_time_unit, mill.dwell_time)
        #
        # # scan_sequence_indices
        #
        # return RasterizedPattern(np.concatenate(scan_lines), out_length_unit, out_time_unit)



"""
Provides the :class:`Transformable` class.
"""
from __future__ import annotations
from typing import Optional, Union, Callable, TypeVar
import abc

from fibomat.linalg.vectors import Vector, VectorLike
from fibomat.linalg.transformables.transformation_builder import _TransformationBuilder
from fibomat.linalg.transformables.transformable_base import TransformableBase
from fibomat.linalg.boundingboxes import BoundingBox


class Transformable(TransformableBase[Vector, float, BoundingBox], abc.ABC):
    """
    :class:`Transformable` is a base class providing the translate, rotate and uniform scale
    transformations.

    In order to use this mixin in a child class, the following methods and properties must be implemented:
        * :attr:`Transformable.center`
        * :meth:`Transformable.translate`
        * :meth:`Transformable.simple_rotate`
        * :meth:`Transformable.simple_scale`

    """

    _VectorClass = Vector

    def __init__(self, description: Optional[str] = None):
        """
        Args:
            pivot (VectorLike, optional): if set, the :attr:`Transformable.pivot` is set to `pivot`. If not set,
                                           :attr:`Transformable.center` is used as default.

            description (str, optional): optional description
        """
        super().__init__(description=description)




from typing import Iterable, Union, Callable, Optional, Tuple, List, Type
import abc

import numpy as np

from fibomat.shapes import HollowArcSpline, ArcSplineCompatible, ArcSpline, Line
from fibomat.linalg import TransformableBase, Vector, DimVector
from fibomat.curve_tools import fill_with_lines
from fibomat.layout.layoutbase import LayoutBase
from fibomat.layout.utils import _round_towards_mean, _check_lattice_vectors
from fibomat.layout.groups.group_base import GroupBase


class LatticeBaseMixin(abc.ABC):
    _VectorType: Type[Union[Vector, DimVector]]

    @staticmethod
    def _gen_lattice_points(
        boundary: Union[HollowArcSpline, ArcSpline], u: Vector, v: Vector, center: Vector, seed: Vector
    ) -> Tuple[np.ndarray, np.ndarray]:
        boundary = boundary.translated_to(center)

        alpha = u.angle_about_x_axis

        # map alpha to [-pi/2, pi/2]
        if alpha < np.pi/2:
            pass
        elif np.pi/2 <= alpha < 1.5*np.pi:
            alpha -= np.pi
        else:
            alpha -= 2*np.pi

        pitch = (v - u.projected(v)).length

        lattice_planes = fill_with_lines(boundary, pitch=pitch, alpha=alpha, invert=False, seed=seed)

        # print(boundary, lattice_planes, seed)

        # translate seed point in first quadrant
        normalized_seed = Vector()

        # s = Sample()
        # s.add_annotation(boundary * U_('µm'))
        # for plane in lattice_planes:
        #     for line in plane:
        #         s.add_annotation(line * U_('µm'))
        #
        # s.plot()

        def lattice_points_on_line(line: Line):
            m = np.array([u, v]).T

            start_lattice_point = np.linalg.solve(m, line.start - center + seed)
            end_lattice_point = np.linalg.solve(m, line.end - center + seed)

            i_u_start, i_u_end = _round_towards_mean(start_lattice_point[0], end_lattice_point[0])

            # TODO: more tests on this!
            if abs(start_lattice_point[0] - end_lattice_point[0]) < 1:
                if not np.isclose(i_u_start, i_u_end):
                    raise ValueError

            # TODO: needed?
            assert np.isclose(start_lattice_point[1], end_lattice_point[1])
            assert np.isclose(start_lattice_point[1], np.rint(start_lattice_point[1]))

            if i_u_start <= i_u_end:
                i_u = np.arange(i_u_start, i_u_end + 1, dtype=int)
            else:
                i_u = np.arange(i_u_end, i_u_start + 1, dtype=int)[::-1]
            i_v = int(np.rint(start_lattice_point[1]))

            return i_u, i_v

        i_u_min = np.iinfo(int).max
        i_u_max = np.iinfo(int).min

        i_v_min = np.iinfo(int).max
        i_v_max = np.iinfo(int).min

        lattice_points = []

        for plane in lattice_planes:
            for line in plane:
                try:
                    points = lattice_points_on_line(line)
                except ValueError:
                    continue
                i_points_u, i_point_v = points

                i_u_min = min(i_u_min, np.min(i_points_u))
                i_u_max = max(i_u_max, np.max(i_points_u))

                i_v_min = min(i_v_min, i_point_v)
                i_v_max = max(i_v_max, i_point_v)

                lattice_points.append(np.c_[i_points_u, [i_point_v]*len(i_points_u)])

        lattice_points_uv = np.concatenate(lattice_points)
        lattice_points_xy = np.outer(lattice_points_uv[:, 0], u) + np.outer(lattice_points_uv[:, 1], v) - seed

        return lattice_points_uv, lattice_points_xy

    @classmethod
    def _generate_impl(
        cls,
        boundary: Union[HollowArcSpline, ArcSplineCompatible],
        u: Vector, v: Vector,
        center: Vector,
        element_gen: Callable,
        predicate: Optional[Union[Callable, List[Callable]]],
        explode: bool, remove_outliers: bool,
        scale_vec: Callable,
        unscale_vec: Callable,
        seed: Vector
    ):
        # check arguments
        if not isinstance(boundary, HollowArcSpline) and not isinstance(boundary, ArcSplineCompatible):
            raise TypeError('boundary must be HollowArcSpline or ArcSplineCompatible')

        if isinstance(boundary, ArcSplineCompatible):
            boundary = boundary.to_arc_spline()

        if not boundary.is_closed:
            raise ValueError('boundary must be closed shaped.')

        _check_lattice_vectors(u, v)

        if predicate is not None:
            if not isinstance(predicate, Iterable):
                predicate = (predicate,)

            if not all([callable(pred) for pred in predicate]):
                raise TypeError('predicate must be Callable or List[Callable].')

        explode = bool(explode)
        remove_outliers = bool(remove_outliers)

        # generate actual lattice site objects
        lattice_points_uv, lattice_points_xy = cls._gen_lattice_points(boundary, u, v, Vector(), seed)

        # shift uv points so that (0, 0) is in the upper left

        # min_u, min_v = np.min(lattice_points_uv, axis=0)

        lattice_points_uv -= np.min(lattice_points_uv, axis=0)

        elements = []
        for lattice_point_xy, lattice_point_uv in zip(lattice_points_xy, lattice_points_uv):
            lattice_site_element = element_gen(unscale_vec(lattice_point_xy), lattice_point_uv)

            if lattice_site_element:
                lattice_site_element = lattice_site_element.translated_to(unscale_vec(lattice_point_xy + center))

                if explode:
                    if isinstance(lattice_site_element, LayoutBase):
                        sub_elements = lattice_site_element._layout_elements()
                        if remove_outliers:
                            for sub_elem in sub_elements:
                                if boundary.contains(scale_vec(sub_elem.pivot)):
                                    elements.append(sub_elem)
                        else:
                            elements.extend(sub_elements)
                    else:
                        elements.append(lattice_site_element)
                else:
                    elements.append(lattice_site_element)

        # sort the points by predicate
        if predicate:
            if explode:
                lattice_elements_xy = np.array([elem.pivot for elem in elements])
            else:
                lattice_elements_xy = lattice_points_xy

            # todo: this looks weird
            sorted_indices = np.lexsort(tuple(pred(lattice_elements_xy) for pred in predicate))

            elements = [elements[i] for i in sorted_indices]

        return elements



import sys

import numpy as np

from vispy import app
from vispy.gloo import context
from vispy.visuals import TextVisual, RectangleVisual
from vispy.visuals.transforms import STTransform


# https://github.com/vispy/vispy/blob/a6ec64bb7b56739b463e3c16245a31940ca4f5f7/vispy/visuals/text/text.py#L151
def _glpyh_dimensions(text, font, anchor_x, anchor_y, lowres_size):
    """Convert text characters to VBO"""
    # Necessary to flush commands before requesting current viewport because
    # There may be a set_viewport command waiting in the queue.
    # TODO: would be nicer if each canvas just remembers and manages its own
    # viewport, rather than relying on the context for this.
    canvas = context.get_current_canvas()
    canvas.context.flush_commands()

    text_vtype = np.dtype([('a_position', np.float32, 2),
                           ('a_texcoord', np.float32, 2)])
    vertices = np.zeros(len(text) * 4, dtype=text_vtype)
    prev = None
    width = height = ascender = descender = 0
    ratio, slop = 1. / font.ratio, font.slop
    x_off = -slop
    # Need to make sure we have a unicode string here (Py2.7 mis-interprets
    # characters like "•" otherwise)
    if sys.version[0] == '2' and isinstance(text, str):
        text = text.decode('utf-8')
    # Need to store the original viewport, because the font[char] will
    # trigger SDF rendering, which changes our viewport
    # todo: get rid of call to glGetParameter!

    # Also analyse chars with large ascender and descender, otherwise the
    # vertical alignment can be very inconsistent
    for char in 'hy':
        glyph = font[char]
        y0 = glyph['offset'][1] * ratio + slop
        y1 = y0 - glyph['size'][1]
        ascender = max(ascender, y0 - slop)
        descender = min(descender, y1 + slop)
        height = max(height, glyph['size'][1] - 2*slop)

    # Get/set the fonts whitespace length and line height (size of this ok?)
    glyph = font[' ']
    spacewidth = glyph['advance'] * ratio
    lineheight = height * 1.5

    # Added escape sequences characters: {unicode:offset,...}
    #   ord('\a') = 7
    #   ord('\b') = 8
    #   ord('\f') = 12
    #   ord('\n') = 10  => linebreak
    #   ord('\r') = 13
    #   ord('\t') = 9   => tab, set equal 4 whitespaces?
    #   ord('\v') = 11  => vertical tab, set equal 4 linebreaks?
    # If text coordinate offset > 0 -> it applies to x-direction
    # If text coordinate offset < 0 -> it applies to y-direction
    esc_seq = {7: 0, 8: 0, 9: -4, 10: 1, 11: 4, 12: 0, 13: 0}

    # Keep track of y_offset to set lines at right position
    y_offset = 0

    # When a line break occur, record the vertices index value
    vi_marker = 0
    ii_offset = 0  # Offset since certain characters won't be drawn

    # The running tracker of characters vertex index
    vi = 0

    orig_viewport = canvas.context.get_viewport()
    for ii, char in enumerate(text):
        if ord(char) in esc_seq:
            if esc_seq[ord(char)] < 0:
                # Add offset in x-direction
                x_off += abs(esc_seq[ord(char)]) * spacewidth
                width += abs(esc_seq[ord(char)]) * spacewidth
            elif esc_seq[ord(char)] > 0:
                # Add offset in y-direction and reset things in x-direction
                dx = dy = 0
                if anchor_x == 'right':
                    dx = -width
                elif anchor_x == 'center':
                    dx = -width / 2.
                vertices['a_position'][vi_marker:vi+4] += (dx, dy)
                vi_marker = vi+4
                ii_offset -= 1
                # Reset variables that affects x-direction positioning
                x_off = -slop
                width = 0
                # Add offset in y-direction
                y_offset += esc_seq[ord(char)] * lineheight
        else:
            # For ordinary characters, normal procedure
            glyph = font[char]
            kerning = glyph['kerning'].get(prev, 0.) * ratio
            x0 = x_off + glyph['offset'][0] * ratio + kerning
            y0 = glyph['offset'][1] * ratio + slop - y_offset
            x1 = x0 + glyph['size'][0]
            y1 = y0 - glyph['size'][1]
            u0, v0, u1, v1 = glyph['texcoords']
            position = [[x0, y0], [x0, y1], [x1, y1], [x1, y0]]
            texcoords = [[u0, v0], [u0, v1], [u1, v1], [u1, v0]]
            vi = (ii + ii_offset) * 4
            vertices['a_position'][vi:vi+4] = position
            vertices['a_texcoord'][vi:vi+4] = texcoords
            x_move = glyph['advance'] * ratio + kerning
            x_off += x_move
            ascender = max(ascender, y0 - slop)
            descender = min(descender, y1 + slop)
            width += x_move
            height = max(height, glyph['size'][1] - 2*slop)
            prev = char

    if orig_viewport is not None:
        canvas.context.set_viewport(*orig_viewport)

    dx = dy = 0
    if anchor_y == 'top':
        dy = -descender
    elif anchor_y in ('center', 'middle'):
        dy = (-descender - ascender) / 2
    elif anchor_y == 'bottom':
        dy = -ascender
    if anchor_x == 'right':
        dx = -width
    elif anchor_x == 'center':
        dx = -width / 2.

    # If any linebreaks occured in text, we only want to translate characters
    # in the last line in text (those after the vi_marker)
    vertices['a_position'][0:vi_marker] += (0, dy)
    vertices['a_position'][vi_marker:] += (dx, dy)
    vertices['a_position'] /= lowres_size

    return width, height


class ScaleBar:
    def __init__(self, unit: str, parent: app.Canvas):
        self._unit = unit

        self._parent = parent

        self._background = RectangleVisual(center=(0, 0), width=20, height=5, color=(1, 1, 1, .75))
        self._background.transform = STTransform(
            translate=(self._parent.physical_size[0] / 2., self._parent.physical_size[1] / 2.)
        )

        self._scale = RectangleVisual(center=(0, 0), width=20, height=5, color=(0,0,0,1))
        self._scale.set_gl_state(blend=True, blend_func=('src_alpha', 'one_minus_src_alpha'))
        self._scale.transform = STTransform(
            translate=(self._parent.physical_size[0] / 2., self._parent.physical_size[1] / 2.)
        )

        self._label = TextVisual('10 µm', bold=False, font_size=12, color='black', pos=(0, 10), anchor_x='center', anchor_y='bottom')
        self._label.transform = STTransform(
            translate=(self._parent.physical_size[0] / 2., self._parent.physical_size[1] / 2.)
        )

    def transform(self, pixel_per_length, scale):
        vp = (0, 0, self._parent.physical_size[0], self._parent.physical_size[1])

        diag = np.sqrt(self._parent.physical_size[0]**2 + self._parent.physical_size[1]**2)

        self._scale.transforms.configure(canvas=self._parent, viewport=vp)
        self._scale.transform.translate=(self._parent.physical_size[0] / 2, self._parent.physical_size[1] / 2)
        self._scale.transform.scale = scale, scale

        self._scale.height = 10 / scale

        scale_bar_length = pixel_per_length * self._parent.physical_size[0] / 5
        if scale_bar_length > 1:
            new_rect_width = 10**int(np.log10(scale_bar_length)) / pixel_per_length / scale
        else:
            new_rect_width = 10**int(np.log10(scale_bar_length) - 1) / pixel_per_length / scale

        self._scale.width = 1 * new_rect_width

        self._label.transforms.configure(canvas=self._parent, viewport=vp)
        self._label.transform.translate=(self._parent.physical_size[0] / 2, self._parent.physical_size[1] / 2)
        self._label.text = f'{pixel_per_length * self._scale.width * scale:.2f} {self._unit}'

        glyph_dims = np.array([
            _glpyh_dimensions(
                t, self._label._font, self._label._anchors[0], self._label._anchors[1], self._label._font._lowres_size
            ) for t in self._label.text
        ])

        self._background.transforms.configure(canvas=self._parent, viewport=vp)
        self._background.transform.translate=(self._parent.physical_size[0] / 2, self._parent.physical_size[1] / 2 + 15)
        self._background.transform.scale = scale, scale

        if self._scale.width > np.sum(glyph_dims[:, 0]) / scale / 2:
            self._background.width = 1.1 * self._scale.width
        else:
            self._background.width = 1.5 * np.sum(glyph_dims[:, 0]) / scale / 2

        self._background.height = 80 / scale

        margin = 10 * scale

        shift_background = (
            self._parent.physical_size[0] - scale * self._background.width / 2 - margin,
            self._parent.physical_size[1] - scale * self._background.height / 2 - margin
        )
        shift = (shift_background[0], shift_background[1] - 15)

        self._background.transform.translate = shift_background
        self._label.transform.translate = shift
        self._scale.transform.translate = shift

    def draw(self):
        self._background.draw()
        self._scale.draw()
        self._label.draw()



"""Provides the `class:`Shape` class."""
from __future__ import annotations
from typing import Optional, Any
import abc

from fibomat.linalg import Transformable
from fibomat.units import U_


class Shape(Transformable, abc.ABC):
    """Baseclass for all shapes in the library which defines the basic interface."""

    def __init__(self, description: Optional[str] = None):
        """
        Args:
            description (str, optional): description
        """
        super().__init__(description=description)

    @abc.abstractmethod
    def __repr__(self) -> str:
        """str: repr of class"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_closed(self) -> bool:
        """
        bool: True if shape is closes. This property should not be defined for 0-dim shapes

        Access:
            get
        """
        raise NotImplementedError

    def __mul__(self, other: Any):
        if isinstance(other, U_):
            from .dim_shape import DimShape
            return DimShape(self, other)
        else:
            raise NotImplementedError



# from fibomat.raster_styles.two_d.contour_parallel import ContourParallel
# from fibomat.raster_styles.two_d.linear import Linear
from fibomat.raster_styles.two_d.linebyline import LineByLine

__all__ = ['ContourParallel', 'Linear', 'LineByLine']



from os import write
from typing import List, Optional, Dict, Any, Union, TextIO
import itertools
import pathlib
import tempfile
import webbrowser

import bokeh.models as bm
import bokeh.plotting as bp
import bokeh.layouts as bl
import bokeh.embed as be
import bokeh.resources as br
from bokeh.embed import file_html
from bokeh.resources import INLINE

import PIL.Image

import numpy as np

from jinja2 import Template

from fibomat.backend import BackendBase
from fibomat.backend.backendbase import shape_type, ShapeNotSupportedError
from fibomat.default_backends._bokeh_site import BokehSite, ShapeType
from fibomat import shapes
from fibomat.linalg import Vector, DimVector, DimVectorLike
from fibomat.site import Site
from fibomat.pattern import Pattern
from fibomat.utils import PathLike
from fibomat.shapes.rasterizedpoints import RasterizedPoints
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, LengthQuantity, U_, Q_, has_length_dim, scale_factor
from fibomat.shapes import Shape, DimShape
from fibomat.mill import Mill
from fibomat.default_backends.measuretool import MeasureTool


from bokeh.resources import JSResources
from bokeh.core.templates import JS_RESOURCES


_old_js_raw = getattr(JSResources, 'js_raw')

here = pathlib.Path(__file__).parent.resolve()
with open(here / 'bokeh-measuretool.min.js') as fp_js:
    _custom_ext_js = fp_js.read()

def _js_raw_patched(self):
    raw = _old_js_raw.fget(self)
    raw.append(_custom_ext_js)
    return raw

setattr(JSResources, 'js_raw', property(_js_raw_patched))

# would be nicer, but impractical because every bokeh template must be fixed..
# maybe set_cache_hook/getcache_hook could be used
# # stolen from https://docs.bokeh.org/en/latest/docs/user_guide/embed.html#standard-template and modified to include measure tool js
# FILE_TEMPLATE = Template("""
# {% from macros import embed %}

# <!DOCTYPE html>
# <html lang="en">
#   {% block head %}
#   <head>
#     {% block inner_head %}
#       <meta charset="utf-8">
#       <title>{% block title %}{{ title | e if title else "Bokeh Plot" }}{% endblock %}</title>
#       {% block preamble %}{% endblock %}
#       {% block resources %}
#         {% block css_resources %}
#           {{ bokeh_css | indent(8) if bokeh_css }}
#         {% endblock %}
#         {% block js_resources %}
#           {{ bokeh_js | indent(8) if bokeh_js }}
#         {% endblock %}
#         <script type="text/javascript">
#         {{ measure_tool | safe | indent(8) }}
#         </script>
#       {% endblock %}
#       {% block postamble %}{% endblock %}
#     {% endblock %}
#   </head>
#   {% endblock %}
#   {% block body %}
#   <body>
#     {% block inner_body %}
#       {% block contents %}
#         {% for doc in docs %}
#           {{ embed(doc) if doc.elementid }}
#           {% for root in doc.roots %}
#             {% block root scoped %}
#               {{ embed(root) | indent(10) }}
#             {% endblock %}
#           {% endfor %}
#         {% endfor %}
#       {% endblock %}
#       {{ plot_script | indent(8) }}
#     {% endblock %}
#   </body>
#   {% endblock %}
# </html>
# """
# )

# class _RenderWrapper:
#     def render(self, *args, **kwargs):
#         here = pathlib.Path(__file__).parent.resolve()
#         with open(here / 'bokeh-measuretool.min.js') as fp_js:
#             custom_js = fp_js.read()

#         return FILE_TEMPLATE.render(*args, measure_tool=custom_js, **kwargs)


class StubRasterStyle(RasterStyle):
    def __init__(self, dimension: int):
        self._dimension = dimension

    @property
    def dimension(self) -> int:
        return self._dimension

    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPoints:
        pass


class BokehImage(DimShape):
    def __init__(self, file: PathLike, image_pixel_scale: LengthQuantity, center: Optional[DimVector] = None):
        self._file = file

        if not has_length_dim(image_pixel_scale):
            raise ValueError('image_pixel_scale must have length dimension')

        self._image_pixel_scale = image_pixel_scale

        image = PIL.Image.open(file).convert('RGBA')

        self._width, self._height = image.size

        self._width *= image_pixel_scale.m
        self._height *= image_pixel_scale.m

        if center is not None:
            self._center = DimVector(center)
        else:
            self._center = DimVector()

        self._image_data = np.asarray(image)

        super().__init__(self, image_pixel_scale.units)

    @property
    def shape(self):
        return self

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def center(self):
        return self._center

    @property
    def data(self):
        return self._image_data

    def _impl_translate(self, trans_vec: DimVectorLike) -> None:
        raise NotImplementedError

    def _impl_rotate(self, theta: float) -> None:
        raise NotImplementedError

    def _impl_scale(self, trans_vec: DimVectorLike) -> None:
        raise NotImplementedError

    def _impl_mirror(self, trans_vec: DimVectorLike) -> None:
        raise NotImplementedError


class BokehBackendBase(BackendBase):
    @shape_type(BokehImage)
    def bokeh_image(self, ptn: Pattern[BokehImage]):
        raise ShapeNotSupportedError


class BokehBackend(BokehBackendBase):
    """
    The default backend for plotting projects, based on the bokeh library.
    All shapes defined in fibomat library are supported.

    .. note:: :class:`~fibomat.shapes.arc.shapes.Arc` and :class:`~fibomat.shapes.curve.shapes.Curve` are rasterized during plotting
              due to lack of supported of the HoverTool for this shapes in the bokeh library. The pitch can be defined
              via the `rasterize_pitch` parameter.
    """
    name = 'bokeh'

    def __init__(
            self, *,
            unit: Optional[LengthUnit] = None,
            title: Optional[str] = None,
            hide_sites: bool = False,
            rasterize_pitch: Optional[LengthQuantity] = None,
            fullscreen: bool = True,
            legend: bool = True,
            cycle_colors: bool = True,
            image_alpha: float = .75,
            **kwargs
    ):
        """
        Args:
            unit (units.UnitType, optional): used unit for plotting, default to units.U_('µm')
            title (str, optional): title of plot, default to ''
            hide_sites (bool, optional): if true, sides' outlines are not shown, default to false
            rasterize_pitch (units.QuantityType. optional): curve_tools.rasterize pitch for shapes.Arc, ... and shapes.Curve, default to units.Q_('0.01 µm')
            fullscreen (bool, optional): if true, plot will be take the whole page, default to True
            cycle_colors (bool): if True, different sites get different colors.
            image_alpha (float): alpha value (transparency) of images, default to 0.75
        """
        super().__init__(**kwargs)

        if unit:
            if not has_length_dim(unit):
                raise ValueError('unit\'s dimension must by [length].')
            self._unit = unit
        else:
            self._unit = U_('µm')

        if title:
            self._title = str(title)
        else:
            self._title = ''

        self._hide_sites = bool(hide_sites)

        if rasterize_pitch:
            if not has_length_dim(rasterize_pitch):
                raise ValueError('rasterize_pitch\'s dimension must by [length].')
            self._rasterize_pitch = rasterize_pitch
        else:
            self._rasterize_pitch = Q_('0.001 µm')

        self._fullscreen = bool(fullscreen)
        self._legend = bool(legend)

        self._cycle_colors = cycle_colors

        self._bokeh_sites: List[BokehSite] = []
        self._annotation_site = BokehSite(
            site_index=-1,
            plot_unit=self._unit,
            dim_center=DimVector(),
            theta=0,
            rasterize_pitch=self._rasterize_pitch,
            description='Annotations'
        )
        self._image_annotation: Optional[BokehImage] = None
        self._image_alpha = float(image_alpha)
        self.fig: Optional[bp.Figure] = None

    def process_site(self, site: Site):
        self._bokeh_sites.append(
            BokehSite(
                site_index=len(self._bokeh_sites),
                plot_unit=self._unit,
                dim_center=site.center,
                theta=site._theta_vec.angle_about_x_axis,
                rasterize_pitch=self._rasterize_pitch,
                cycle_colors=self._cycle_colors,
                dim_fov=site.fov,
                description=site.description
            )
        )
        super().process_site(site)

    def process_pattern(self, ptn: Pattern) -> None:
        super().process_pattern(ptn)

    def process_unknown(self, ptn: Pattern) -> None:
        bbox = ptn.dim_shape.shape.bounding_box
        bbox_ptn = Pattern(
            dim_shape=shapes.Rect(bbox.width, bbox.height, 0, bbox.center) * ptn.dim_shape.unit,
            mill=ptn.mill,
            raster_style=ptn.raster_style,
            **ptn.kwargs  # True if 'annotation' in ptn.kwargs else False
        )
        self._filled_curve(bbox_ptn)

    def _collect_plot_data(self, shape_type: ShapeType) -> Dict[str, Any]:
        # https://stackoverflow.com/a/40826547
        keys = BokehSite.plot_data_keys
        data_dicts = [site.plot_data[shape_type] for site in self._bokeh_sites]
        data_dicts.append(self._annotation_site.plot_data[shape_type])
        return {key: list(itertools.chain(*[data_dict[key] for data_dict in data_dicts])) for key in keys}

    @staticmethod
    def _create_datasources(extra_keys: Optional[List[str]] = None):
        keys = BokehSite.plot_data_keys if not extra_keys else BokehSite.plot_data_keys + extra_keys

        def _create():
            return bm.ColumnDataSource({key: [] for key in keys})

        return {'spots': _create(), 'non_filled': _create(), 'filled': _create()}



    @staticmethod
    def _create_renderers(fig, data_sources):
        spot_glyphs = fig.circle_x(
            x='x', y='y',
            fill_color='color', line_color='color', fill_alpha=.25,
            legend_group='site_id',
            size=10,
            source=data_sources['spots']
        )

        non_filled_curve_glyphs = fig.multi_line(
            xs='x', ys='y',
            line_color='color', line_width=2,
            legend_group='site_id',
            source=data_sources['non_filled']
        )

        filled_curve_glyphs = fig.multi_polygons(
            xs='x', ys='y',
            line_width=2,
            fill_color='color', line_color='color', fill_alpha='fill_alpha',
            hatch_pattern='hatch_pattern',
            legend_group='site_id',
            source=data_sources['filled']

        )

        return {'spots': spot_glyphs, 'non_filled': non_filled_curve_glyphs, 'filled': filled_curve_glyphs}

    def _plot_impl(self, data_sources):
        spot_data = self._collect_plot_data(ShapeType.SPOT)
        non_filled = self._collect_plot_data(ShapeType.NON_FILLED_CURVE)
        filled = self._collect_plot_data(ShapeType.FILLED_CURVE)

        data_sources['spots'].stream(spot_data)
        data_sources['non_filled'].stream(non_filled)
        data_sources['filled'].stream(filled)

        return {
            'spots': len(spot_data['x']),
            'non_filled': len(non_filled['x']),
            'filled': len(filled['x'])
        }


    def plot(self):
        tooltips = [
            # ('type', 'shape'),
            ('shape', '@shape_prop'),
            # ('collection_index', '@collection_index'),
            ('mill', '@mill'),
            ('raster style', '@raster_style'),
            ('site', '@site_id'),
            ('description', '@description')
            # ('mill_settings', '@mill_settings'),
        ]

        site_tooltips = [
            # ('site', '@site'),
            ('description', '@description')
        ]

        fig = bp.figure(
            title=self._title,
            x_axis_label=f'x / {self._unit:~P}', y_axis_label=f'y / {self._unit:~P}',
            match_aspect=True,
            sizing_mode='stretch_both' if self._fullscreen else 'stretch_width',
            tools="pan,wheel_zoom,reset,save"
        )

        fig.add_tools(
            MeasureTool(
                measure_unit=f'{self._unit:~P}',  # line_color=bc.groups.red.Crimson, line_width=3  # bpal.all_palettes['Colorblind'][4][3]
            )
        )

        fig.add_tools(bm.BoxZoomTool(match_aspect=True))

        data_sources = self._create_datasources()
        renderers = self._create_renderers(fig, data_sources)

        self._plot_impl(data_sources)

        # spot_glyphs = fig.circle_x(
        #     x='x', y='y',
        #     fill_color='color', line_color='color', fill_alpha=.25,
        #     legend_group='site_id',
        #     size=10,
        #     source=bm.ColumnDataSource(self._collect_plot_data(ShapeType.SPOT))
        # )

        # non_filled_curve_glyphs = fig.multi_line(
        #     xs='x', ys='y',
        #     line_color='color', line_width=2,
        #     legend_group='site_id',
        #     source=bm.ColumnDataSource(self._collect_plot_data(ShapeType.NON_FILLED_CURVE))
        # )

        # filled_curve_glyphs = fig.multi_polygons(
        #     xs='x', ys='y',
        #     line_width=2,
        #     fill_color='color', line_color='color', fill_alpha='fill_alpha',
        #     hatch_pattern='hatch_pattern',
        #     legend_group='site_id',
        #     source=bm.ColumnDataSource(self._collect_plot_data(ShapeType.FILLED_CURVE))
        # )

        # images
        if self._image_annotation:
            image = self._image_annotation

            center = image.center.vector_as(self._unit)

            input_offset_x = bm.Spinner(title="Image x", low=-100000, high=100000, step=0.01, value=center.x, width=80)
            input_offset_y = bm.Spinner(title="Image y", low=-100000, high=100000, step=0.01, value=center.y, width=80)

            image_center = image.center.vector_as(self._unit)
            image_scale = scale_factor(self._unit, image.unit)
            width = image.width * image_scale
            height = image.height * image_scale
            
            image_data = image.data.view(dtype=np.uint32).reshape(image.data.shape[:2])[::-1,:]
 
            rendered_image = fig.image_rgba(
                image=[image_data],
                x=image_center.x - width/2,
                y=image_center.y - width/2,
                dw=width,
                dh=height,
                # anchor="center",
                global_alpha=self._image_alpha
            )

            input_offset_x.js_link('value', rendered_image.glyph, 'x')
            input_offset_y.js_link('value', rendered_image.glyph, 'y')

        # layers
        # https://github.com/bokeh/bokeh/issues/9087
        if not self._hide_sites:
            site_glyphs = fig.multi_polygons(
                xs='x', ys='y',
                line_width=2,
                fill_color='color', line_color='color', fill_alpha='fill_alpha', line_alpha='fill_alpha',
                legend_group='site_id',
                source=bm.ColumnDataSource(self._collect_plot_data(ShapeType.SITE))
            )

            site_glyphs_hover = bm.HoverTool(
              renderers=[site_glyphs],
              tooltips=site_tooltips,
              point_policy='follow_mouse'
            )
            fig.add_tools(site_glyphs_hover)

        # hover tool for shapes
        # add shape hover tool after site hovertool so it is rendered on top of the site tooltip
        shape_glyphs_hover = bm.HoverTool(
            # renderers=[
            #     spot_glyphs, non_filled_curve_glyphs,  filled_curve_glyphs
            # ],
            renderers=list(renderers.values()),
            tooltips=tooltips, point_policy='follow_mouse'
        )
        fig.add_tools(shape_glyphs_hover)

        def sorter(item):
            value = item.label['value']
            if value == 'Annotations':
                return -1
            else:
                return int(value.split(',')[0].split(' ')[1])

        legend_tmp = {x.label['value']: x for x in fig.legend.items}.values()
        fig.legend.items.clear()
        fig.legend.items.extend(sorted(
            legend_tmp,
            key=sorter
        ))

        fig.legend.visible = self._legend

        if self._image_annotation:
            self.fig = bl.column([bl.row([input_offset_x, input_offset_y]), fig], width_policy='max')
        else:
            self.fig = fig

    # def _gen_html(self, use_cdn: bool):
    #     resources = br.CDN if use_cdn else br.INLINE
    #     return be.file_html(models=self.fig, resources=resources, title=self._title, template=_RenderWrapper())

    def show(self):
        # with tempfile.NamedTemporaryFile('w', suffix='.html', delete=False) as fp:
        #     fp.write(self._gen_html(True))
        #     fp.flush()
        #     webbrowser.open(fp.name)
        bp.show(self.fig)

    def save(self, filename: PathLike, use_cdn: bool = False):       
        # with open(filename, 'w') as fp:
        #     fp.write(self._gen_html(use_cdn))
        resources = br.CDN if use_cdn else br.INLINE
        bp.save(self.fig, filename, resources=resources)

    def spot(self, ptn: Pattern[shapes.Spot]) -> None:
        if '_annotation' in ptn.kwargs:
            self._annotation_site.spot(ptn)
        else:
            self._bokeh_sites[-1].spot(ptn)

    def _non_filled_curve(self, ptn):
        if '_annotation' in ptn.kwargs:
            self._annotation_site.non_filled_curve(ptn)
        else:
            self._bokeh_sites[-1].non_filled_curve(ptn)

    def _filled_curve(self, ptn):
        if '_annotation' in ptn.kwargs:
            self._annotation_site.filled_curve(ptn)
        else:
            self._bokeh_sites[-1].filled_curve(ptn)

    def _filled_curve_with_holes(self, ptn):
        if '_annotation' in ptn.kwargs:
            self._annotation_site.filled_curve_with_holes(ptn)
        else:
            self._bokeh_sites[-1].filled_curve_with_holes(ptn)

    def _dispatch_pattern(self, ptn):
        if not ptn.dim_shape.shape.is_closed:
            self._non_filled_curve(ptn)
        elif isinstance(ptn.dim_shape.shape, shapes.HollowArcSpline):
            self._filled_curve_with_holes(ptn)
        elif ptn.raster_style.dimension < 2:
            self._non_filled_curve(ptn)
        else:
            self._filled_curve(ptn)

    def line(self, ptn: Pattern[shapes.Line]) -> None:
        self._dispatch_pattern(ptn)

    def polyline(self, ptn: Pattern[shapes.Polyline]) -> None:
        self._dispatch_pattern(ptn)

    def arc(self, ptn: Pattern[shapes.Arc]) -> None:
        self._dispatch_pattern(ptn)

    def arc_spline(self, ptn: Pattern[shapes.ArcSpline]) -> None:
        self._dispatch_pattern(ptn)

    def polygon(self, ptn: Pattern[shapes.Polygon]) -> None:
        self._dispatch_pattern(ptn)

    def rect(self, ptn: Pattern[shapes.Rect]) -> None:
        self._dispatch_pattern(ptn)

    def ellipse(self, ptn: Pattern[shapes.Ellipse]) -> None:
        self._dispatch_pattern(ptn)

    def circle(self, ptn: Pattern[shapes.Circle]) -> None:
        self._dispatch_pattern(ptn)

    def rasterized_points(self, ptn: Pattern[shapes.RasterizedPoints]):
        rect = shapes.Rect.from_bounding_box(ptn.dim_shape.shape.bounding_box)

        new_pattern = Pattern(
            dim_shape=rect * ptn.dim_shape.unit,
            mill=ptn.mill,
            raster_style=ptn.raster_style,
            description=ptn.description,
            **ptn.kwargs
        )

        if '_annotation' in ptn.kwargs:
            self._annotation_site.filled_curve(new_pattern, hatch_pattern='/')
        else:
            self._bokeh_sites[-1].filled_curve(new_pattern, hatch_pattern='/')

    def hollow_arc_spline(self, ptn: Pattern[shapes.HollowArcSpline]) -> None:
        self._dispatch_pattern(ptn)

    def bokeh_image(self, ptn: Pattern[BokehImage]):
        if '_annotation' in ptn.kwargs:
            if not self._image_annotation:
                self._image_annotation = ptn.dim_shape
            else:
                raise RuntimeError('currently, only one image is supported.')
        else:
            raise RuntimeError('BokehImage can only added to annotation site')



"""This submodule implements (physical) unit support for the package (units = physical units).

It uses the python module `pint`. For usage, see the example below or the
documentation at https://pint.readthedocs.io/en/0.10.1/ ::

    from fibomat.units import Q_, U_

    # factor of 1. exists implicitly
    a = 1. * Q_('µm') = Q_('1. µm') = 1. * U_('µm')

    a_in_m = a.to('m')

    a_raw = a.magnitude
    a_unit = a.units

    dose = 12 * Q('ions / nm**2')
    dose_in_c_per_um2 = dose.to('coulomb / µm**2')
"""
from typing import Union, NewType

import pint  # type: ignore


ureg: pint.registry.UnitRegistry = pint.UnitRegistry(system='mks')
"""pint unit registry"""

# add unit `Ions` to the unit registry
# we are only interested in charges +1, so it is fine
ureg.define('ions = elementary_charge')  # = ions ?

# define doses as derived dimensions
ureg.define('[spotdose] = [current] * [time]')
ureg.define('[linedose] = [current] * [time] / [length]')
ureg.define('[areadose] = [current] * [time] / [length] ** 2')


Q_ = ureg.Quantity
"""Quantity constructor."""


U_ = ureg.Unit
"""Unit constructor"""


UnitType = Union[pint.Unit]  # NewType('UnitType', pint.Unit)
"""Generic pint unit type used for typing annotation. Objects constructed with :attr:`U_` hae this type."""


# https://github.com/python/mypy/issues/6701
# class PintUnit:
#     def __getattr__(self, item: str) -> Any:
#         pass


LengthUnit = pint.Unit  # NewType('LengthUnit', UnitType)
TimeUnit = pint.Unit  # NewType('TimeUnit', UnitType)
# LengthUnit = NewType('LengthUnit', UnitType)


QuantityType = Union[pint.Quantity]  # NewType('QuantityType', pint.Quantity)
"""pint quantity type used for typing annotation"""

LengthQuantity = QuantityType  # NewType('LengthQuantity', QuantityType)
TimeQuantity = QuantityType  # NewType('TimeQuantity', QuantityType)


def has_length_dim(quant_or_unit: Union[UnitType, QuantityType]) -> bool:
    """Check if `quant_or_unit` has dimension `[length]`.

    Args:
        quant_or_unit (UnitType, QuantityType): unit or quantity to be checked.

    Returns:
        bool
    """
    return quant_or_unit.dimensionality == '[length]'
    # return ureg.check(qunat, '')


def has_time_dim(quant_or_unit: Union[UnitType, QuantityType]) -> bool:
    """Check if `quant_or_unit` has dimension `[time]`.

    Args:
        quant_or_unit (UnitType, QuantityType): unit or quantity to be checked.

    Returns:
        bool
    """
    return quant_or_unit.dimensionality == '[time]'


def scale_factor(base_unit: Union[UnitType, QuantityType], other_unit: Union[UnitType, QuantityType]) -> float:
    """Calculates the scaling factor needed to convert `other_unit` to `base_untit`.

    .. warning:: If `base_unit` or `other_unit` are quantities, the magnitude is ignored.

    Args:
        base_unit (Union[UnitType, QuantityType]): base unit
        other_unit (Union[UnitType, QuantityType]): other unit

    Returns:
        float: scaling factor
    """
    base = base_unit if isinstance(base_unit, pint.Unit) else 1. * base_unit.units
    other_quant = 1. * other_unit if isinstance(other_unit, pint.Unit) else 1. * other_unit.units

    return other_quant.to(base).magnitude


def scale_to(base_unit: Union[UnitType, QuantityType], quantity: QuantityType) -> float:
    """Scales quantity `quantity` to `base_unit` and returns the result as `float`.

    Args:
        base_unit (Union[UnitType, QuantityType]): base unit
        quantity (QuantityType): quantity to be scaled

    Returns:
        float: scaled quantity
    """
    base = base_unit if isinstance(base_unit, pint.Unit) else base_unit.units
    return quantity.to(base).magnitude



import sys
import argparse
import configparser

import numpy as np

from PyQt5 import QtWidgets

from fibomat.linalg import BoundingBox
from fibomat.beam_simulation.window import AnimationWindow


def _load_and_prepare(file: str):

    with open(file, 'r') as fp:
        info_lines = []
        while 'Points' not in (line := fp.readline()):
            info_lines.append(line)

        points = np.loadtxt(fp, delimiter=' ')

    info = configparser.ConfigParser()
    info.read_string(''.join(info_lines))
    length_unit = info['Info']['length_unit']

    # points[:, 2] = 0.
    bbox = BoundingBox.from_points(points[:, :2])
    points[:, :2] -= bbox.center

    # flip at x axis because the canvas system is defined like this !?
    points[:, 1] *= -1

    return points, length_unit


def run():
    parser = argparse.ArgumentParser(description='Ion beam simulation.')
    parser.add_argument('filename', metavar='filename', type=str, nargs=1,
                        help='filename of spot list file')

    args = parser.parse_args()
    file = args.filename[0]

    appQt = QtWidgets.QApplication(sys.argv)
    win = AnimationWindow(*_load_and_prepare(file))
    win.show()
    appQt.exec_()



from __future__ import annotations
from typing import Iterable, Any

import numpy as np

from fibomat.linalg.boundingboxes.boundingbox_base import BoundingBoxBase
from fibomat.linalg.vectors import Vector, VectorLike
from fibomat.units import U_


class BoundingBox(BoundingBoxBase[Vector, float]):
    _VectorClass = Vector

    @classmethod
    def from_points(cls, points: Iterable[VectorLike]) -> BoundingBox:
        """
        Constructs rectangular bounding box containing all `points`

        Args:
            points (VectorArrayLike): points which should be included in bounding box

        Returns:
            (BoundingBox): new `BoundingBox`
        """
        points = np.asarray(points)

        if len(points) == 0 or points.ndim != 2 or points.shape[1] != 2:
            raise ValueError('points must not be empty and contain vector compatible values.')

        return BoundingBox(np.min(points, axis=0), np.max(points, axis=0))

    def __mul__(self, other: Any):
        if isinstance(other, U_):
            from .dim_boundingbox import DimBoundingBox
            return DimBoundingBox(self._lower_left * other, self._upper_right * other)
        else:
            raise NotImplementedError





from fibomat.backend.backendbase import BackendBase


class DoNothingBackend(BackendBase):
    """This backend is only for testing purpose."""

    name = 'donothing'

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        super().__init__()

    def null_shape(self, shape, mill, shape_units, **kwargs):
        self.null_shape = (shape, mill, shape_units, kwargs)  # noqa




from fibomat.curve_tools.combine import combine_curves
from fibomat.curve_tools.intersections import self_intersections, curve_intersections
from fibomat.curve_tools.offset import offset, offset_with_islands, deflate, inflate
from fibomat.curve_tools.rasterize import rasterize, fill_with_lines, rasterize_with_const_error
from fibomat.curve_tools.biarc_approximation import approximate_parametric_curve
from fibomat.curve_tools.smooth import smooth

__all__ = [
    'combine_curves', 'self_intersections', 'curve_intersections', 'offset', 'offset_with_islands', 'rasterize',
    'approximate_parametric_curve', 'deflate', 'inflate', 'fill_with_lines', 'smooth'
]



from typing import Sequence, Any

import numpy as np

from fibomat.linalg.vectors.vector_base import VectorBase, FloatTypes, SelfT
from fibomat.linalg.vectors.vector import Vector
from fibomat.units import Q_, U_, scale_factor


class DimVector(VectorBase[Q_]):
    @staticmethod
    def _validate_type(val: Any) -> bool:
        return isinstance(val, Q_) and val.u.dimensionality == '[length]' and isinstance(val.m, FloatTypes)

    @staticmethod
    def _build_np_array(val: Sequence[Q_]) -> np.ndarray:
        x: Q_ = val[0]
        y: Q_ = val[1]

        y.ito(x.u)

        return np.array([x.m, y.m], dtype=float) * x.u

    @staticmethod
    def _make_zero_vector():
        return [0*U_('µm'), 0*U_('µm')]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def vector(self) -> Vector:
        return Vector(self._vector.m)

    @property
    def unit(self) -> U_:
        return self._vector.u

    def normalized(self: SelfT) -> SelfT:
        """Create a new vector with same :attr:`Vector.phi` but :attr:`Vector.r` = 1.

        Returns:
            SelfT
        """
        return self.__class__(self._vector / np.linalg.norm(self._vector.m))

    def vector_as(self, unit: U_) -> Vector:
        """Returns a scaled vector in units of `unit`.

        Args:
            unit:

        Returns:
            Vector
        """
        return scale_factor(unit, self._vector.u) * Vector(self._vector.m)



"""
Provides the :class:`Pattern` class.
"""
# pylint: disable=protected-access
from __future__ import annotations
from typing import Generic, TypeVar, Optional, Dict

from fibomat.mill import MillBase
from fibomat.linalg import DimTransformable, DimVector, DimVectorLike, DimBoundingBox
from fibomat.raster_styles import RasterStyle
from fibomat.shapes import DimShape


DimTransformableT = TypeVar('DimTransformableT', bound=DimTransformable)


class Pattern(DimTransformable, Generic[DimTransformableT]):
    """
    Class is used to collect a shape with a length unit,  mill settings and optional settings.
    """

    def __init__(
        self,
        dim_shape: DimShape,
        mill: MillBase,
        raster_style: RasterStyle,
        *,
        description: Optional[str] = None,
        **kwargs
    ):
        """

        Args:
            dim_shape (Tuple[ShapeType, units.LengthUnit]):
                tuple of a shape type and its length unit. ShapeType can be any transformable, e.g. a layout.Group or
                shapes.Line, ...
            mill (Mill): mill object
            **kwargs: additional args
        """
        super().__init__(description=description)

        self._dim_shape = dim_shape
        self._mill = mill
        self._raster_style = raster_style
        self._kwargs = kwargs

    @property
    def dim_shape(self) -> DimShape:
        """Dimensioned shape of the pattern.

        Access:
            get

        Returns:
            DimShape
        """
        return self._dim_shape

    @property
    def mill(self) -> MillBase:
        """Mill of the pattern.

        Access:
            get

        Returns:
            MillBase
        """
        return self._mill

    @property
    def raster_style(self) -> RasterStyle:
        """Raster style of the pattern.

        Access:
            get

        Returns:
            RasterStyle
        """
        return self._raster_style

    @property
    def kwargs(self) -> Dict:
        """Additional passed arguments.

        Access:
            get

        Returns:
            Dict
        """
        return self._kwargs

    @property
    def bounding_box(self) -> DimBoundingBox:
        return self.dim_shape.bounding_box

    @property
    def center(self) -> DimVector:
        return self._dim_shape.center

    def _impl_translate(self, trans_vec: DimVectorLike) -> None:
        self._dim_shape._impl_translate(DimVector(trans_vec))

    def _impl_rotate(self, theta: float) -> None:
        self._dim_shape._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        self._dim_shape._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: DimVectorLike) -> None:
        self._dim_shape._impl_mirror(DimVector(mirror_axis))



"""Provides the :class:`Ellipse` class."""
# pylint: disable=too-many-arguments,invalid-name

from __future__ import annotations
from typing import Optional

import numpy as np

from fibomat.shapes.shape import Shape
from fibomat.linalg import VectorLike, Vector, BoundingBox, rotate, translate
from fibomat.shapes.arc_spline import ArcSpline, ArcSplineCompatible
from fibomat.shapes.circle import Circle


class Ellipse(Shape, ArcSplineCompatible):
    """2-dim ellipse."""
    def __init__(
        self,
        a: float, b: float, theta: float = 0,
        center: Optional[VectorLike] = None,
        description: Optional[str] = None
    ):
        """

        Args:
            a (float): length of half axis in pos. x-direction (unrotated)
            b (float): length of half axis in pos. y-direction (unrotated)
            theta (float): rotation angle, default to 0
            center (VectorLike, optional): center of circle, default to (0, 0)
            description (str, optional): description
        """
        super().__init__(description)

        # raise NotImplementedError

        assert a > 0.
        assert b > 0.

        # center, axis in pos. x direction, axis in pos. y direction (all unrotated)
        # self._axes = VectorArray(
        #     (float(a), 0.),
        #     (0., float(b))
        # ).rotated(float(theta))

        self._norm_a_axes = Vector(1, 0.).rotated(float(theta))

        self._a = float(a)
        self._b = float(b)

        self._center = Vector(center) if center is not None else Vector()

    def __repr__(self) -> str:
        return '{}(a={!r}, b={!r}, theta={!r}, center={!r})'.format(
            self.__class__.__name__, self.a, self.b, self.theta, self.center)

    def to_arc_spline(self) -> ArcSpline:
        if np.isclose(self._a, self._b):
            return Circle(r=self._a, center=self._center).to_arc_spline()

        from fibomat.shapes.parametric_curve import ParametricCurve

        from sympy import Curve
        from sympy.abc import t
        from sympy import sin, cos

        ellipse = ParametricCurve.from_sympy_curve(
            Curve([self.a * cos(t), self.b * sin(t)], (t, 0, 2*np.pi)),
            try_length_integration=False
        )

        ellipse_arc_spline = ellipse.to_arc_spline()

        return ellipse_arc_spline.transformed(rotate(self.theta) | translate(self.center))

    @property
    def a(self) -> float:
        """Length of half axis in pos. x-direction (unrotated)

        Access:
            get

        Returns:
            float
        """
        return self._a

    @property
    def b(self) -> float:
        """Length of half axis in pos. y-direction (unrotated)

        Access:
            get

        Returns:
            float
        """
        return self._b

    @property
    def theta(self) -> float:
        """rotation angle of ellipse.

        Access:
            get

        Returns:
            float
        """
        return self._norm_a_axes.angle_about_x_axis

    @property
    def center(self) -> Vector:
        return self._center

    @property
    def bounding_box(self) -> BoundingBox:
        # TODO: test this carefully!
        # https://stackoverflow.com/questions/87734/how-do-you-calculate-the-axis-aligned-bounding-box-of-an-ellipse
        # https://www.iquilezles.org/www/articles/ellipses/ellipses.htm
        u = np.array(self._a * self._norm_a_axes)
        # b axis os orthogoan to a axis
        v = np.array((self._b * self._norm_a_axes).rotated(self.theta + np.pi/4))

        w = np.sqrt(u*u + v*v)
        return BoundingBox(self._center-w, self._center+w)

    @property
    def is_closed(self) -> bool:
        return True

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._center += Vector(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        self._axes = self._norm_a_axes.rotated(float(theta))

    def _impl_scale(self, fac: float) -> None:
        self._center *= float(fac)
        self._a *= float(fac)
        self._b *= float(fac)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._norm_a_axes = self._norm_a_axes.mirrored(mirror_axis)
        self._center = self._center.mirrored(mirror_axis)
        # TODO: change a and b??



from typing import List

import svgelements

from fibomat.shapes import Shape
from fibomat.utils import PathLike


from typing import List, Optional, Callable
import warnings
import itertools

import numpy as np

import svgelements

import anytree

from fibomat.units import Q_, U_
from fibomat.shapes import Shape, Rect, Circle, Polyline, Polygon, Polyline, Line, Ellipse
from fibomat.layout import Group
from fibomat.linalg import mirror, translate, Vector, BoundingBox
from fibomat.utils import PathLike

# def path_type(path: svgelements.Path):
#     for segment in path:
#         if not isinstance(segment, svgelements.MO):

class IndexedSubpath:
    def __init__(self, subpath, index):
        self.subpath = subpath
        self.index = index


# def make_bbox(path: IndexedSubpath) -> BoundingBox:
#     # box: xmin, ymin, xmax, ymax
#     box = path.subpath.bbox()
#     print(box)
#     return BoundingBox(box[:2], box[2:])



# def buildpath_tree_impl(paths: List[IndexedSubpath], _root_node, root_bbox, i_node):
#     path = paths.pop(0)
#     path_bbox = make_bbox(path)


#     if path_bbox in root_bbox:
#         pass
#     pass 


# def sort_by_bboxes(paths: List[IndexedSubpath]):
#     def _get_max_bbox(paths_):


#     bboxes = [make_bbox(path) for path in paths]

#     i_sorted = []



# def build_path_tree(paths: List[IndexedSubpath]):
#     # we assume that the paths are hierarchically ordered (meaning that a subpath which lies completely in a nother one comes after the path it lies in) 

#     class PathData:
#         def __init__(self, /, **kwargs):
#             self.__dict__.update(**kwargs)

#     if not paths:
#         return

#     root = anytree.Node('root')

#     nodes = []
#     for i, path in enumerate(paths):
#         nodes.append(anytree.Node(i, parent=root, path_data=PathData(path=path.subpath, bbox=make_bbox(path), index=path.index)))

#     nodes.sort(key=lambda node: -node.path_data.bbox.area)

#     print([node.path_data.bbox for node in nodes])
    
#     def build_subtree(root_node, nodes):
#         while nodes:
#             possible_child = nodes.pop(0)
#             if possible_child.path_data.bbox in root_node.path_data.bbox:
#                 possible_child.parent = root_node
#             else:
#                 possible_child.parent = root_node.parent
#                 build_subtree(possible_child, nodes)
    
#     # print(current_root)

#     build_subtree(nodes[0], nodes[1:])

#     # while nodes:
#     #     current_root = nodes[0]
#     #     print('current root', current_root.name)

#     #     for possible_child in nodes[1:]:
#     #         if possible_child.path_data.bbox in current_root.path_data.bbox:


#     # # for _, _, node in anytree.RenderTree(root): 
#     # for node in nodes[1:]:
#     #     print('current root', current_root.name)
#     #     print('node', node.name, node.path_data)
#     #     print(node.path_data.bbox in current_root.path_data.bbox)

#     #     if node.path_data.bbox in current_root.path_data.bbox:
#     #         print(f'node {node.name} ind {current_root.name}')
#     #         node.parent = current_root
#     #         current_root = node
#     #     else:
#     #         # current_root = current_root.parent
#     #         print('searching new root, old root is ', current_root.name)
#     #         while node.path_data.bbox in current_root.path_data.bbox and current_root.parent.name != 'root':
#     #             current_root = current_root.parent

#     #             print('new root', current_root)
    
#     for pre, fill, node in anytree.RenderTree(root):
#         print("%s%s" % (pre, node.name))
#     # current_root_path = paths.pop(0)

#     # anytree.Node(i_node, parent=_root, segment=current_root_path)
#     # i_node += 1


#     # current_root_bbox = make_bbox(current_root_path)

    
#     #     if make_bbox(path) in current_root_bbox:
#     #         pass

#     # return _root

#     return None


def path_is_closed(path: svgelements.Subpath):
    segments = path.segments()
    start = segments[0].end
    end = segments[-1].end

    return np.allclose(start, end)


def is_polyline_or_polygon(path: svgelements.Subpath):
    polyline_segment_types = svgelements.Move, svgelements.Line, svgelements.Close
    for segment in path.segments():
        if not isinstance(segment, polyline_segment_types):
            return False
    
    return True


def categorize(elements: List[IndexedSubpath], condition: Callable):
    first, second = [], []
    for element in elements:
        (first, second)[not condition(element.subpath)].append(element)

    return first, second
    

def parse_polyline(polyline_subpaths: IndexedSubpath):
    points = []

    segments = list(polyline_subpaths.subpath.segments())

    for segment in segments[1:]:
        start = segment.start
        # end = segment.end
        # print(start)
        points.append(Vector(*start))
    
    points.append(Vector(*segments[-1].end))

    return IndexedSubpath(Polyline(points), polyline_subpaths.index)


def parse_polygon(polygon_subpaths: IndexedSubpath):
    points = []

    segments = list(polygon_subpaths.subpath.segments())

    for segment in segments[1:]:
        start = segment.start
        # end = segment.end
        # print(start)
        points.append(Vector(*start))

    print(segments)

    return IndexedSubpath(Polygon(points), polygon_subpaths.index)


def parse_path(path: svgelements.Path):
    # https://stackoverflow.com/a/12135169

    subpaths = [IndexedSubpath(subpath, i) for i, subpath in enumerate(path.as_subpaths())]

    closed_paths, open_paths = categorize(subpaths, path_is_closed)
    

    polylines, non_polyline_open_paths = categorize(open_paths, is_polyline_or_polygon)

    polygons, non_polygon_closed_paths = categorize(closed_paths, is_polyline_or_polygon)

    if non_polyline_open_paths:
        warn_element_not_supported('non_polyline_open_paths')

    if non_polygon_closed_paths:
        warn_element_not_supported('non_polygon_closed_paths')

    fibomat_polygons = [parse_polygon(polygon) for polygon in polygons]

    fibomat_polylines = [parse_polyline(polyline) for polyline in polylines]
    # build_path_tree(polygons)
    
    
    # [ for polygon in polygons]

    # print(polylines)
    # print(open_curves)
    # print(closed_paths)

    to_be_chained = []

    if fibomat_polygons:
        print(fibomat_polygons)

        to_be_chained.append(fibomat_polygons)

    if fibomat_polylines:
        print(fibomat_polylines)
        to_be_chained.append(fibomat_polylines)

    merged = list(itertools.chain(
        *to_be_chained
        # open_curves
    ))

    print('merged', merged)

    if merged:
        return Group(list(map(
            lambda index_path: index_path.subpath,
            sorted(merged, key=lambda index_path: index_path.index)
        )))

    # raise RuntimeError



def is_reified(element: svgelements.SVGElement) -> bool:
    def matrix_as_array(m):
        return np.array([m.a, m.c, m.b, m.d, m.e, m.f], dtype=np.float64)
    unit_matrix = np.array([1., 0., 0., 1., 0., 0.], dtype=np.float64)

    return np.allclose(matrix_as_array(element.transform), unit_matrix)


def warn_element_not_supported(element: svgelements.SVGElement, reason: Optional[str] = None):
    msg = f'Unsupported shape of type "{type(element)}"'
    if reason:
        msg += f' (reason: {reason})'
    
    msg += '. Try to convert it to path in your favorite svg editor first. For now, it will be skipped.'
    
    warnings.warn(msg)


def parse_elements(elements: List[svgelements.SVGElement]) -> List:
    parsed_shapes = []

    for element in elements:
        try:
            if element.values['visibility'] == 'hidden':
                continue
        except (KeyError, AttributeError):
            pass
        
        if isinstance(element, svgelements.SVG):
            continue
        elif isinstance(element, svgelements.Path):
            if len(element) != 0:
                reified_path = element.reify()
                if paths_group := parse_path(reified_path):
                    parsed_shapes.append(paths_group)
        elif isinstance(element, svgelements.Shape):
            reified_shape = element.reify()

            if is_reified(reified_shape):
                if isinstance(reified_shape, svgelements.Rect):
                    rect: svgelements.Rect = reified_shape
                    if reified_shape.rx > 0 or reified_shape.ry > 0:
                        warn_element_not_supported(reified_shape, 'rounded corners')
                    else:
                        parsed_shapes.append(Rect(
                            width=rect.width, height=rect.height, center=(rect.x + rect.width / 2, rect.y + rect.height / 2)
                        ))
                elif isinstance(reified_shape, svgelements.Circle):
                    circle: svgelements.Circle = reified_shape
                    parsed_shapes.append(Circle(
                            r=circle.r, center=(circle.cx, circle.cy)
                    ))
                elif isinstance(reified_shape, svgelements.Ellipse):
                    ellipse: svgelements.Ellipse = reified_shape
                    parsed_shapes.append(Ellipse(
                            a=ellipse.rx, b=ellipse.ry, center=(ellipse.cx, ellipse.cy)
                    ))
                elif isinstance(reified_shape, svgelements.Line):
                    line: svgelements.Line = reified_shape
                    parsed_shapes.append(Line(
                            start=line.start, end=line.end
                    ))
                elif isinstance(reified_shape, (svgelements.Polyline, svgelements.Polygon)):
                    parsed_shapes.append(parse_path(reified_shape))
                else:
                    warn_element_not_supported(reified_shape)
            else:
                # path can be always reified
                reified_path = svgelements.Path(element).reify()
                if paths_group := parse_path(reified_path):
                    parsed_shapes.append(paths_group)
        elif isinstance(element, svgelements.Group):
            if group_elements := parse_elements(element):
                parsed_shapes.append(
                    Group(group_elements)
                )
        else:
            # raise RuntimeError(f'Unsupported shape "{element}".')
            warn_element_not_supported(element)

    return parsed_shapes


def shapes_from_svg(file_path: PathLike, scale: Q_) -> Optional[Group]:
    """[summary]

    Args:
        file_path (PathLike): [description]
        scale (Q_): defines the length reference. E.g.: if scale = 1cm, 1cm in the svg file will be 1 length unit in the import shapes.

    Returns:
        Group: [description]
    """

    scale = scale.m_as('inch')

    svg = svgelements.SVG.parse(file_path, ppi=1/scale)

    if parsed_svg := parse_elements(svg):
        return Group(parsed_svg).transformed(mirror((1, 0)) | translate((-svg.width / 2, svg.height / 2)))


from typing import Dict, List, Any, Optional, Union
import copy
import enum

import numpy as np

import bokeh.palettes as bp

from fibomat.pattern import Pattern
from fibomat import shapes
from fibomat.curve_tools import rasterize

from fibomat.linalg import Vector, DimVector, DimVectorLike
from fibomat.units import LengthUnit, scale_factor, LengthQuantity, scale_to


@enum.unique
class ShapeType(enum.Enum):
    SPOT = 'spot'
    NON_FILLED_CURVE = 'non_filled_curve'
    FILLED_CURVE = 'filled_curve'
    SITE = 'site'


def _css_to_rgb(css: str):
    css = css[1:]
    return tuple(int(css[2*i:2*(i+1)], 16) for i in range(3))


_colors = {
    'shape': bp.all_palettes['Colorblind'][4][1],
    'shape_alpha': .5,
    'site': bp.all_palettes['Colorblind'][4][0],
    'site_alpha': 0.25,
    'other': bp.all_palettes['Colorblind'][4][2],
    'other_alpha': .25
}


class BokehSite:
    plot_data_keys: List[str] = [
        'x', 'y', 'x_hole', 'y_hole',
        'site_id', 'shape_type', 'shape_prop', 'raster_style', 'mill',
        'color', 'fill_alpha', 'hatch_pattern', 'description'
    ]

    def __init__(
        self,
        site_index: int,
        plot_unit: LengthUnit,
        dim_center: DimVectorLike,
        theta: float,
        rasterize_pitch: LengthQuantity,
        cycle_colors: bool = False,
        dim_fov: Optional[DimVector] = None,
        description: Optional[str] = None,
    ):
        shape_types = [ShapeType.SPOT, ShapeType.NON_FILLED_CURVE, ShapeType.FILLED_CURVE, ShapeType.SITE]
        plot_data_proto = {key: [] for key in self.plot_data_keys}
        self.plot_data = {shape_type: copy.deepcopy(plot_data_proto) for shape_type in shape_types}

        self._plot_unit = plot_unit

        dim_center = DimVector(dim_center)
        self._center: Vector = scale_factor(plot_unit, dim_center.unit) * dim_center.vector

        self._theta = theta

        if dim_fov:
            dim_fov = DimVector(dim_fov)
            self._fov: Vector = scale_factor(plot_unit, dim_fov.unit) * dim_fov.vector
        else:
            self._fov = None

        self._rasterize_pitch = rasterize_pitch

        self._colors = copy.deepcopy(_colors)
        if cycle_colors:
            # color = RGB(*_css_to_rgb(bp.YlGnBu9[site_index % len(bp.YlGnBu9)]))
            color = bp.Viridis11[site_index % len(bp.YlGnBu9)]
            self._colors['site'] = color
            self._colors['shape'] = color

        if self._fov is not None:
            self._site_description = f'Site {site_index}, fov=({self._fov.x}, {self._fov.y}, center=({self._center.x}, {self._center.y})) {self._plot_unit:~P}'
            self._site_description += (', ' + str(description)) if description else ''
            self._label = self._site_description

            points = np.array(shapes.Rect(width=self._fov.x, height=self._fov.y, theta=self._theta, center=self._center).corners)

            self._add_plot_data(
                shape_type=ShapeType.SITE,
                x=[[list(points[:, 0])]], y=[[list(points[:, 1])]],
                site_id=str(self._label), shape_prop='Site',
                raster_style='', mill='',
                color=self._colors['site'], fill_alpha=self._colors['site_alpha'],
                description=self._site_description
            )
        else:
            # it's a annotation layer
            self._site_description = 'Annotations'
            self._label: str = self._site_description
            self._colors['shape'] = 'black'  # bp.all_palettes['Colorblind'][4][3]

    def _add_plot_data(self, shape_type: ShapeType, **kwargs):
        for key in self.plot_data_keys:
            try:
                self.plot_data[shape_type][key].append(kwargs[key])
            except KeyError:
                self.plot_data[shape_type][key].append(None)

    def spot(self, ptn: Pattern[shapes.Spot]) -> None:
        s = scale_factor(self._plot_unit, ptn.dim_shape.unit)
        shape: shapes.Spot = ptn.dim_shape.shape

        color = ptn.kwargs.get('_color', self._colors['shape'])
        if not color:
            color = self._colors['shape']

        self._add_plot_data(
            shape_type=ShapeType.SPOT,
            x=self._center.x + s * shape.center.x, y=self._center.y + s * shape.center.y,
            site_id=str(self._label), shape_prop=str(shape),
            raster_style=str(ptn.raster_style), mill=str(ptn.mill),
            color=color,
            description=f'Pattern: {ptn.description} | Shape: {ptn.dim_shape.shape.description}',
        )

    def _segmentize_pattern(self, ptn: Pattern) -> np.ndarray:
        curve = shapes.ArcSpline.from_shape(ptn.dim_shape.shape)
        s = scale_factor(self._plot_unit, ptn.dim_shape.unit)

        points = []

        for segment in curve.segments:
            if isinstance(segment, shapes.Line):
                segment: shapes.Line
                points.append([np.asarray(segment.start)])
            elif isinstance(segment, shapes.Arc):
                segment: shapes.Arc
                # set the rasterizing pitch so that the maximum distance between rasterized polygon and real arc is
                # self._rasterize_pitch
                delta = scale_to(ptn.dim_shape.unit, self._rasterize_pitch)
                pitch_squared = 4 * (segment.radius**2 - (segment.radius - delta)**2)
                if pitch_squared > 0:
                    arc_points = rasterize(
                        shapes.ArcSpline.from_shape(segment), np.sqrt(pitch_squared)
                    ).dwell_points[:, :2]
                    points.append(arc_points)
                else:
                    raise RuntimeError(
                        'Could not rasterize curve with given rasterize_pitch. Try to decrease rasterize_pitch.'
                    )
            else:
                raise RuntimeError

        if curve.segments:
            points.append([np.asarray(curve.segments[-1].end)])

        points = np.concatenate(points).reshape(-1, 2)
        points *= s
        points += self._center

        return points

    def non_filled_curve(self, ptn: Pattern):
        points = self._segmentize_pattern(ptn)

        color = ptn.kwargs.get('_color', self._colors['shape'])
        if not color:
            color = self._colors['shape']

        self._add_plot_data(
            shape_type=ShapeType.NON_FILLED_CURVE,
            x=list(points[:, 0]), y=list(points[:, 1]),
            site_id=str(self._label), shape_prop=str(ptn.dim_shape.shape),
            raster_style=str(ptn.raster_style), mill=str(ptn.mill),
            color=color,
            description=f'Pattern: {ptn.description} | Shape: {ptn.dim_shape.shape.description}',
        )

    def filled_curve(self, ptn: Pattern, hatch_pattern: Optional[str] = None):
        points = self._segmentize_pattern(ptn)

        color = ptn.kwargs.get('_color', self._colors['shape'])
        if not color:
            color = self._colors['shape']

        self._add_plot_data(
            shape_type=ShapeType.FILLED_CURVE,
            x=[[list(points[:, 0])]], y=[[list(points[:, 1])]],
            site_id=str(self._label), shape_prop=str(ptn.dim_shape.shape),
            raster_style=str(ptn.raster_style), mill=str(ptn.mill),
            color=color, fill_alpha=self._colors['shape_alpha'],
            hatch_pattern=hatch_pattern,
            description=f'Pattern: {ptn.description} | Shape: {ptn.dim_shape.shape.description}',
        )

    def filled_curve_with_holes(self, ptn: Pattern[shapes.HollowArcSpline], hatch_pattern: Optional[str] = None):
        boundary_points = self._segmentize_pattern(Pattern(
            dim_shape=ptn.dim_shape.shape.boundary * ptn.dim_shape.unit, mill=None, raster_style=None
        ))

        hole_points_x = []
        hole_points_y = []

        for hole in ptn.dim_shape.shape.holes:
            points = self._segmentize_pattern(Pattern(
                dim_shape=hole * ptn.dim_shape.unit, mill=None, raster_style=None
            ))
            hole_points_x.append(list(points[:, 0]))
            hole_points_y.append(list(points[:, 1]))

        color = ptn.kwargs.get('_color', self._colors['shape'])
        if not color:
            color = self._colors['shape']

        self._add_plot_data(
            shape_type=ShapeType.FILLED_CURVE,
            x=[[list(boundary_points[:, 0]), *hole_points_x]], y=[[list(boundary_points[:, 1]), *hole_points_y]],
            site_id=str(self._label), shape_prop=str(ptn.dim_shape.shape),
            raster_style=str(ptn.raster_style), mill=str(ptn.mill),
            color=color, fill_alpha=self._colors['shape_alpha'],
            hatch_pattern=hatch_pattern,
            description=f'Pattern: {ptn.description} | Shape: {ptn.dim_shape.shape.description}',
        )



"""Provide rasterization routines."""
from typing import List, Tuple, Optional, Dict, Any, Union

import numpy as np

from fibomat.shapes.arc import Arc
from fibomat.shapes.line import Line
from fibomat.shapes.polygon import Polygon
from fibomat.shapes.polyline import Polyline
from fibomat.shapes.rasterizedpoints import RasterizedPoints
from fibomat.shapes.shape import Shape
from fibomat.shapes.arc_spline import ArcSplineCompatible, ArcSpline
from fibomat.shapes.hollow_arc_spline import HollowArcSpline
from fibomat.linalg import Vector, translate, rotate, VectorLike, BoundingBox
from fibomat.curve_tools.intersections import curve_intersections
from fibomat.shapes._line_non_continuous import LineNonContinuous
from fibomat import _libfibomat


def _rasterize_arc_spline_non_continuous_curve(
    curve: Union[ArcSpline, LineNonContinuous], pitch: float
) -> RasterizedPoints:
    # pylint: disable=invalid-name,too-many-locals

    pitch = float(pitch)
    # n_points = int(curve.length / pitch) + 1

    points = []

    i_points = 0
    offset = 0.
    for segment in curve.segments:
        if isinstance(segment, Arc):
            arc: Arc = segment

            offset_angle = offset / arc.radius
            pitch_angle = pitch / arc.radius

            if offset_angle <= arc.theta:
                points_on_arc = int((arc.theta - offset_angle) / pitch_angle) + 1

                if np.isclose((arc.theta - offset_angle) / pitch_angle - points_on_arc - 1, 0.):
                    points_on_arc += 1

                if not arc.sweep_dir:
                    pitch_angle *= -1.
                    arc_start_angle = arc.start_angle - offset_angle
                else:
                    arc_start_angle = arc.start_angle + offset_angle

                center = np.array(arc.center)
                theta = pitch_angle*np.arange(points_on_arc)

                points.append(
                    center + arc.radius * np.column_stack(
                        (np.cos(theta + arc_start_angle), np.sin(theta + arc_start_angle))
                    )
                )

                i_points += points_on_arc
                offset = pitch - (arc.length - offset - (len(theta)-1) * pitch)
            else:
                offset -= arc.length

        elif isinstance(segment, Line):
            line: Line = segment
            if offset <= line.length:
                direction = (line.end - line.start).normalized()

                start = line.start + direction * offset

                points_on_line = int((line.length-offset) / pitch) + 1

                direction = np.array(direction)
                start = np.array(start)

                t = pitch * np.arange(points_on_line)
                # points[i_points:i_points+points_on_line, :2] = start + np.repeat(t[None, :], 2, axis=0).T * direction
                points.append(start + np.repeat(t[None, :], 2, axis=0).T * direction)
                # points[i_points:i_points+points_on_line, 2] = 1.

                i_points += points_on_line
                offset = pitch - (line.length - t[-1] - offset)
            else:
                offset -= line.length
        else:
            raise RuntimeError(f'Cannot rasterize segment type {segment.__class__.__name__}')

    # assert i_points == n_points
    dwell_points = np.ones(shape=(i_points, 3))
    dwell_points[:, :2] = np.concatenate(points)

    return RasterizedPoints(dwell_points, curve.is_closed)


def rasterize(curve: Shape, pitch: float) -> RasterizedPoints:
    """Rasterize the outline of a Shape with a given pitch uniformly.

    For this, the shape must be convertible to an ArcSpline or must have a `rasterize` method expecting the pitch as
    input.

    Args:
        curve (Shape): curve
        pitch (float): pitch (spacing of points)

    Returns:
        RasterizedPoints

    Raises:
        ValueError: Raised if curve is no ArcSpline, ArcSplineCompatible or not have a `rasterize` method.
    """
    if isinstance(curve, ArcSpline) or isinstance(curve, LineNonContinuous):
        return _rasterize_arc_spline_non_continuous_curve(curve, pitch)

    if raster_method := getattr(curve, 'rasterize', None):
        return raster_method(curve, pitch)

    if isinstance(curve, ArcSplineCompatible):
        return _rasterize_arc_spline_non_continuous_curve(curve.to_arc_spline(), pitch)

    raise ValueError(f'Cannot rasterize the passed object of type {curve.__class__}.')


def rasterize_with_const_error(curve: ArcSpline, error: float) -> Union[Polyline, Polygon]:
    """Convert a arc spline to polygon or polyline.

    Args:
        curve: arc spline
        error: maximum distance between curve an returned polyine/polygon

    Returns:
        Polygon if arc spline is closed and polyline otherwise
    """

    approx = ArcSpline(_libfibomat.convert_arcs_to_lines(curve.arc_spline_impl, error))

    if curve.is_closed:
        return Polygon(approx.vertices[:, :2])
    else:
        return Polyline(approx.vertices[:, :2])


def _make_line(
    curve: ArcSpline, intersection_interval: Tuple[Dict[str, Any], Dict[str, Any]], holes: List[ArcSpline]
) -> Optional[Line]:
    start = Vector(intersection_interval[0]['pos'])
    end = Vector(intersection_interval[1]['pos'])

    assert np.isclose(start.y, end.y)
    assert end.x > start.x

    midpoint = Vector((start.x + end.x) / 2, start.y)

    if curve.contains(midpoint):
        for hole in holes:
            if hole.contains(midpoint):
                return
        return Line(start, end)


def fill_with_lines(
    shape: Union[ArcSpline, HollowArcSpline],
    pitch: float,
    alpha: float,
    invert: bool,
    seed: Optional[VectorLike] = None
) -> List[List[Line]]:
    """Fill a closed shape with lines which are rotated by `alpha`-

    Args:
        shape (Union[ArcSpline, HollowArcSpline]): closed curve to be filled
        pitch (float): distance between lines
        alpha (float): rotation angle of lines with respect to x-axis
        invert (bool):


    Returns:
        List[ArcSpline]

    Raises:
        ValueError: Raised if angle < 0 or angle > pi.
        NotImplementedError: Raised if some kind of singularities occur. This can be fixed (probably) if the shape is
                             rotated.
    """
    # pylint: disable=invalid-name,too-many-locals

    # TODO: cache spatial tree for intersection calculations

    if isinstance(shape, ArcSpline):
        curve = shape
        if not curve.is_closed:
            raise ValueError('ArcSpline is not closed.')
        holes = []
    elif isinstance(shape, HollowArcSpline):
        curve = shape.boundary
        holes = shape.holes
    else:
        raise TypeError(f'Shape must be ArcSpline or HollowArcSpline (got {type(shape)}).')

    if not (-np.pi/2 <= alpha <= np.pi/2):
        raise ValueError('alpha < -pi/2 or alpha > pi/2')

    seed = (Vector(seed) - curve.center).rotated(-alpha) if seed is not None else Vector()
    # seed = Vector(seed) if seed is not None else Vector()

    center = curve.center

    curve = curve.transformed(
        translate(-center) | rotate(-alpha)
    )

    holes = list(map(lambda hole: hole.transformed(translate(-center) | rotate(-alpha)), holes))

    extend = 2 * pitch  # 2 * pitch
    bbox_curve = curve.bounding_box
    bbox = BoundingBox(bbox_curve.lower_left - (extend, extend), bbox_curve.upper_right + (extend, extend))

    bbox_left = bbox.lower_left.x
    bbox_right = bbox.upper_right.x
    bbox_top = pitch * (np.fix(bbox.upper_right.y / pitch)) + seed.y - pitch * np.fix(seed.y / pitch) # pitch * np.fix((bbox.upper_right.y - seed.y) / pitch)
    bbox_bottom = bbox_top - bbox.height # pitch * np.fix((bbox.center.y - bbox.height / 2) / pitch)

    height = np.abs(bbox_top - bbox_bottom)

    y = np.arange(-height / 2, height / 2, pitch)

    ny = len(y)

    if invert:
        intersection_line_base = ArcSpline([(bbox_left, bbox_bottom, 0), (bbox_right, bbox_bottom, 0)], False)
        pitch = -pitch
    else:
        intersection_line_base = ArcSpline([(bbox_left, bbox_top, 0), (bbox_right, bbox_top, 0)], False)

    fill_lines: List[List[Line]] = []

    for i_y in range(ny):
        intersection_line = intersection_line_base.translated((0, -i_y*pitch))

        intersections = curve_intersections(curve, intersection_line)['intersections']

        # ignore coincidences for now
        # # TODO: handle case, if cl_inter contains coincidence
        # assert len(cl_intersections['coincidences']) == 0

        if len(intersections) > 1:
            for hole in holes:
                intersections.extend(curve_intersections(hole, intersection_line)['intersections'])

            intersections_sorted = sorted(
                intersections,
                key=lambda intersection: intersection['pos'][0]
            )

            fill_line = []

            for intersection_interval in zip(intersections_sorted, intersections_sorted[1:]):
                if part_fill_line := _make_line(curve, intersection_interval, holes):
                    fill_line.append(
                        part_fill_line.transformed(
                            translate(center) | rotate(alpha, origin=center)
                        )
                    )

            if fill_line:
                fill_lines.append(fill_line)

    return fill_lines



import numpy as np

from fibomat.units import LengthUnit, TimeUnit, has_length_dim, has_time_dim


class RasterizedPattern:
    """Class to represent dwell points (position + dwell time). In contrast to RasterizedPoints, actual dwell times
    are collected in this class."""

    def __init__(self, dwell_points: np.ndarray, length_unit: LengthUnit, time_unit: TimeUnit):
        """
        Args:
            dwell_points (np.ndarray):
                list of dwell pints with shape (n, 3). each point should contain (x, y, dwell_time).
            length_unit (LengthUnit): length unit of points
            time_unit (TimeUnit): time unit of dwell times
        """
        if dwell_points.ndim != 2 or dwell_points.shape[1] != 3:
            raise ValueError('dwell_points must be a list of shape (n, 3).')

        if not has_length_dim(length_unit):
            raise ValueError('lenght_unit must have dimension [length]')

        if not has_time_dim(time_unit):
            raise ValueError('lenght_unit must have dimension [time]')

        self._dwell_points = dwell_points
        self._length_unit = length_unit
        self._time_unit = time_unit

    @property
    def dwell_points(self):
        """List of dwell points (position + dwell time)

        Returns:
            np.ndarray

        Access:
            get
        """
        view = self._dwell_points[:]
        view.flags.writeable = False
        return view

    @property
    def positions(self):
        """List of positions

        Returns:
            np.ndarray

        Access:
            get
        """
        view = self._dwell_points[:, :2]
        view.flags.writeable = False
        return view

    @property
    def dwell_times(self):
        """List of dwell times

        Returns:
            np.ndarray

        Access:
            get
        """
        view = self._dwell_points[:, 2]
        view.flags.writeable = False
        return view

    @property
    def length_unit(self):
        """Length unit of positions

        Returns:
            LengthUnit

        Access:
            get
        """
        return self._length_unit

    @property
    def time_unit(self):
        """Time unit of dwell times

        Returns:
            TimeUnit

        Access:
            get
        """
        return self._time_unit



"""Provides the :class:`Polygon` class."""
from typing import Optional, List

import numpy as np

from fibomat.shapes import polyline
from fibomat.linalg import VectorLike, Vector


class Polygon(polyline.Polyline):
    """2-dim polygon (closed polyline)."""

    def __init__(self, points: List[VectorLike], description: Optional[str] = None):
        """
        .. note:: `point[0]` and `point[-1]` are automatically connected. Hence, `point[0] != point[-1]` usually.

        Args:
            points (VectorArrayLike): polygon points.
            description (str, optional): description
        """
        super().__init__(points, description, _closed=True)

    @classmethod
    def regular_ngon(  # pylint: disable=too-many-arguments,invalid-name
            cls,
            n: int,
            radius: float = 1.,
            circumcircle: bool = True,
            center: Optional[VectorLike] = None,
            description: Optional[str] = None
    ):
        """
        Creates an regular polygon.

        Args:
            n (int): number of corners
            radius: cirumcircle or incircle radius
            circumcircle (bool): if true, radius is treated as circumcircle radius else as incircle radius
            center (VectorLike, optional): center of ngon, default to (0, 0)
            description (str, optional): description

        Returns:
            Polygon

        Raises:
            ValueError: Raised of n < 3 or radius <= 0.
        """
        if n < 3:
            raise ValueError('n < 3')

        if radius <= 0.:
            raise ValueError('radius <= 0.')

        if not circumcircle:
            radius = radius / np.cos(np.pi / n)

        center = Vector(center) if center is not None else Vector(0, 0)

        angle = 2*np.pi / n

        points = [Vector(radius, 0.)]
        for i in range(1, n):
            points.append(points[i-1].rotated(angle))

        return cls(np.array(points) + center, description)

    @property
    def is_closed(self) -> bool:
        return True



from typing import Union, Tuple

import numpy as np

from fibomat.shapes.line import Line
from fibomat.shapes.arc import Arc
from fibomat.shapes.arc_spline import ArcSpline
from fibomat.curve_tools.intersections import curve_intersections
from fibomat.curve_tools.offset import offset_with_islands, offset
from fibomat.linalg import angle_between, Vector
from fibomat.linalg.helpers import GeomLine
from fibomat.utils.math import mod_2pi



class NonSmoothableError(RuntimeError):
    """"""


def make_tangent_vector(start: np.ndarray, end: np.ndarray):
    tangent = np.asarray(end - start, dtype=float)
    tangent /= np.linalg.norm(tangent)

    return tangent


def make_normal_vector(other: np.ndarray):
    normal = np.array((-other[1], other[0]))

    return normal


def intersection_on_arc(arc: Arc, intersection: np.ndarray):
    intersection_angle = Vector(intersection - arc.center).angle_about_x_axis

    norm_intersection_angle = mod_2pi(intersection_angle - arc.start_angle)
    norm_end_arc_angle = mod_2pi(arc.end_angle - arc.start_angle)

    if not arc.sweep_dir:
        return 0. <= norm_end_arc_angle <= norm_intersection_angle
    else:
        return 0. <= norm_intersection_angle <= norm_end_arc_angle


def intersection_on_line(line: Line, intersection: np.ndarray):
    geom_line = GeomLine(direction=line.end-line.start, support=line.start)

    return 0. <= geom_line.find_param(intersection) <= 1.


def arc_arc_intersection(arc_0: Arc, arc_1: Arc):
    c_0, r_0 = arc_0.center, arc_0.radius
    c_1, r_1 = arc_1.center, arc_1.radius

    c_0 = np.array(c_0)
    c_1 = np.array(c_1)

    d = np.linalg.norm(c_0 - c_1)

    if d > r_0 + r_1:
        raise RuntimeError('No intersection (d > r_0 + r_1)')
    elif d < abs(r_0 - r_1):
        raise RuntimeError('No intersection (d < abs(r_0 - r_1))')
    elif np.isclose(d, 0.) and np.isclose(r_0, r_1):
        raise RuntimeError('No intersection (arcs are identical)')
    else:
        a = (r_0**2 - r_1**2 + d**2) / (2 * d)

        h = np.sqrt(r_0**2 - a**2)

        c = c_0 + a * (c_1 - c_0) / d

        p_1 = np.array([
            c[0] + h * (c_1[1] - c_0[1]) / d,
            c[1] - h * (c_1[0] - c_0[0]) / d
        ])

        p_2 = np.array([
            c[0] - h * (c_1[1] - c_0[1]) / d,
            c[1] + h * (c_1[0] - c_0[0]) / d
        ])

        if np.allclose(p_1, p_2):
            intersection = p_1
        else:
            if intersection_on_arc(arc_0, p_1) and intersection_on_arc(arc_1, p_1):
                intersection = p_1
            elif intersection_on_arc(arc_0, p_2) and intersection_on_arc(arc_1, p_2):
                intersection = p_2
            else:
                raise NonSmoothableError

        return intersection


# http://paulbourke.net/geometry/circlesphere/
def arc_line_intersection(arc: Arc, line: Line):
    x_1 = line.start.x
    x_2 = line.end.x
    x_3 = arc.center.x

    y_1 = line.start.y
    y_2 = line.end.y
    y_3 = arc.center.y

    r = arc.radius

    a = (x_2 - x_1)**2 + (y_2 - y_1)**2
    b = 2 * ((x_2 - x_1) * (x_1 - x_3) + (y_2 - y_1) * (y_1 - y_3))
    c = x_3**2 + y_3**2 + x_1**2 + y_1**2 - 2 * (x_3*x_1 + y_3*y_1) - r**2

    discr = b**2 - 4 * a * c

    if discr < 0:
        raise RuntimeError
    elif np.isclose(discr, 0):
        raise RuntimeError
    else:
        u_1 = (-b + np.sqrt(discr)) / (2*a)
        u_2 = (-b - np.sqrt(discr)) / (2*a)

        p_1 = line.start + u_1 * (line.end - line.start)
        p_2 = line.start + u_2 * (line.end - line.start)

        if intersection_on_arc(arc, p_1) and intersection_on_line(line, p_1):  #
            intersection = p_1
        elif intersection_on_arc(arc, p_2) and intersection_on_line(line, p_2):  #
            intersection = p_2
        else:
            raise NonSmoothableError

        return intersection


def line_line_intersection(line_1: Line, line_2: Line):
    geom_line_1 = GeomLine.make_bisector(line_1.start, line_1.end)
    geom_line_2 = GeomLine.make_bisector(line_2.start, line_2.end)

    intersection = geom_line_1.intersect_at(geom_line_2)

    if not intersection_on_line(line_1, intersection) or not intersection_on_line(line_2, intersection):
        raise NonSmoothableError

    return geom_line_1.intersect_at(geom_line_2)


def make_arc_func(segment: Arc, other_tangent: np.array, kink: np.ndarray, radius: float):
    phi_0 = segment.start_angle
    arc_dir = -1 if not segment.sweep_dir else 1
    theta = segment.theta
    arc_center = segment.center
    arc_radius = segment.radius

    normal_vec = make_normal_vector(segment.unit_tangent_at(0.25 * segment.theta))

    if np.dot(normal_vec, other_tangent) < 0:
        normal_vec *= -1

    if np.dot(normal_vec, kink - segment.center) > 0:
        arc_normal_dir = 1
    else:
        arc_normal_dir = -1

    # def arc_func(t_):
    #     phi = arc_dir * t_ * theta + phi_0
    #     return (
    #         arc_center
    #         + arc_radius * np.array([np.cos(phi), np.sin(phi)])
    #         - arc_normal_dir * radius * np.array([-np.cos(phi), -np.sin(phi)])
    #     )
    #
    # def arc_param(t_):
    #     phi = arc_dir * t_ * theta + phi_0
    #     return (
    #         arc_center
    #         + arc_radius * np.array([np.cos(phi), np.sin(phi)])
    #     )

    arc_offset_seg = Arc(
        radius=arc_normal_dir * radius + segment.radius,
        start_angle=segment.start_angle,
        end_angle=segment.end_angle,
        sweep_dir=segment.sweep_dir,
        center=segment.center
    )

    # return arc_func, arc_param, arc_offset_seg
    return arc_offset_seg


def make_segments(
    left_vertex: np.ndarray, kink_vertex: np.ndarray, right_vertex: np.ndarray, radius: float
):
    left_bulge = left_vertex[2]
    right_bulge = kink_vertex[2]

    kink = np.asarray(kink_vertex[:2], dtype=float)
    left = np.asarray(left_vertex[:2], dtype=float)
    right = np.asarray(right_vertex[:2], dtype=float)

    # build curve segments. the left segment has an inverted direction
    if np.isclose(left_bulge, 0.):
        left_segment = Line(start=kink, end=left)
        left_tangent_start = make_tangent_vector(start=kink, end=left)
    else:
        left_segment = Arc.from_bulge(start=kink, end=left, bulge=-left_bulge)
        left_tangent_start = left_segment.unit_tangent_start

    if np.isclose(right_bulge, 0.):
        right_segment = Line(start=kink, end=right)
        right_tangent_start = make_tangent_vector(start=kink, end=right)
    else:
        right_segment = Arc.from_bulge(start=kink, end=right, bulge=right_bulge)
        right_tangent_start = right_segment.unit_tangent_start

    left_normal_vec = make_normal_vector(left_tangent_start)
    right_normal_vec = make_normal_vector(right_tangent_start)

    if isinstance(left_segment, Line):
        if np.dot(left_normal_vec, right_tangent_start) < 0:
            left_normal_dir = -1
        else:
            left_normal_dir = 1

        left_offset = Line(
            start=left_segment.start + radius * left_normal_dir * left_normal_vec,
            end=left_segment.end + radius * left_normal_dir * left_normal_vec
        )

    else:
        # left_func, left_param,
        left_offset = make_arc_func(left_segment, right_tangent_start, kink, radius)

    if isinstance(right_segment, Line):
        if np.dot(right_normal_vec, left_tangent_start) < 0:
            right_normal_dir = -1
        else:
            right_normal_dir = 1

        right_offset = Line(
            start=right_segment.start + radius * right_normal_dir * right_normal_vec,
            end=right_segment.end + radius * right_normal_dir * right_normal_vec
        )

    else:
        # right_func, right_param,
        right_offset = make_arc_func(right_segment, left_tangent_start, kink, radius)

    return (left_segment, right_segment), (left_offset, right_offset)


def make_smoothing_arc(
    left_segment: Union[Line, Arc],
    left_offset: Union[Line, Arc],
    right_segment: Union[Line, Arc],
    right_offset: Union[Line, Arc],
):
    if isinstance(left_segment, Arc) and isinstance(right_segment, Arc):
        smoothing_arc_center = arc_arc_intersection(left_offset, right_offset)
        smoothing_arc_start = Vector(smoothing_arc_center - left_segment.center).normalized_to(left_segment.radius) + left_segment.center
        smoothing_arc_end = Vector(smoothing_arc_center - right_segment.center).normalized_to(right_segment.radius) + right_segment.center
    elif isinstance(left_segment, Line) and isinstance(right_segment, Line):
        smoothing_arc_center = line_line_intersection(left_offset, right_offset)

        # https://en.wikibooks.org/wiki/Linear_Algebra/Orthogonal_Projection_Onto_a_Line
        s1 = left_segment.end - left_segment.start
        v1 = smoothing_arc_center - left_segment.start
        smoothing_arc_start = s1.dot(v1) / s1.dot(s1) * s1 + left_segment.start
        s2 = right_segment.end - right_segment.start
        v2 = smoothing_arc_center - right_segment.start
        smoothing_arc_end = s2.dot(v2) / s2.dot(s2) * s2 + right_segment.start
    elif isinstance(left_segment, Line) and isinstance(right_segment, Arc):
        smoothing_arc_center = arc_line_intersection(arc=right_offset, line=left_offset)
        s1 = left_segment.end - left_segment.start
        v1 = smoothing_arc_center - left_segment.start
        smoothing_arc_start = s1.dot(v1) / s1.dot(s1) * s1 + left_segment.start
        smoothing_arc_end = Vector(smoothing_arc_center - right_segment.center).normalized_to(right_segment.radius) + right_segment.center
    elif isinstance(left_segment, Arc) and isinstance(right_segment, Line):
        smoothing_arc_center = arc_line_intersection(arc=left_offset, line=right_offset)
        smoothing_arc_start = Vector(smoothing_arc_center - left_segment.center).normalized_to(left_segment.radius) + left_segment.center
        s2 = right_segment.end - right_segment.start
        v2 = smoothing_arc_center - right_segment.start
        smoothing_arc_end = s2.dot(v2) / s2.dot(s2) * s2 + right_segment.start
    else:
        raise RuntimeError

    return smoothing_arc_start, smoothing_arc_center, smoothing_arc_end


def make_smoothed_vertices(
    smoothing_arc_points: Tuple,
    left_vertex, kink_vertex, right_vertex,
    left_segment, right_segment
):
    smoothing_arc_start, smoothing_arc_center, smoothing_arc_end = smoothing_arc_points

    if isinstance(left_segment, Line):
        new_left_vertex = left_vertex
    else:
        new_left_vertex = (
            *left_vertex[:2],
            np.tan(
                angle_between(
                    left_segment.center - smoothing_arc_start,
                    left_segment.center - left_vertex[:2]
                ) / 4
            ) * np.sign(left_vertex[2])
        )

    if isinstance(right_segment, Line):
        new_right_vertex = (*smoothing_arc_end, 0)
    else:
        new_right_vertex = (
            *smoothing_arc_end,
            np.tan(
                angle_between(
                    right_segment.center - smoothing_arc_end,
                    right_segment.center - right_vertex[:2]
                ) / 4
            ) * np.sign(kink_vertex[2])
        )

    sweep_matrix = np.ones(shape=(3, 3), dtype=float)
    sweep_matrix[0, 1:] = np.asarray(smoothing_arc_start)
    sweep_matrix[1, 1:] = np.asarray(kink_vertex[:2])
    sweep_matrix[2, 1:] = np.asarray(smoothing_arc_end)

    smooth_arc_sweep_dir = 1 if np.linalg.det(sweep_matrix) > 0 else -1

    new_arc = (
        *smoothing_arc_start,
        np.tan(
            angle_between(
                smoothing_arc_center - smoothing_arc_start,
                smoothing_arc_center - smoothing_arc_end
            ) / 4
        ) * smooth_arc_sweep_dir
    )

    return new_left_vertex, new_arc, new_right_vertex


def smooth(arc_spline: ArcSpline, radius: float):
    kinks = arc_spline.kinks()

    if kinks:
        vertices = list(arc_spline.vertices)

        wrap = lambda index: index % len(vertices)

        index_offset = 0

        for i_kink in kinks:
            i_kink_with_offset = i_kink + index_offset
            # print(i_kink, vertices[i_last_kink:i_kink-1])
            kink_vertex = vertices[wrap(i_kink_with_offset)]

            left_vertex = vertices[wrap(i_kink_with_offset-1)]
            right_vertex = vertices[wrap(i_kink_with_offset+1)]

            (left_segment, right_segment), (left_offset, right_offset) = make_segments(
                left_vertex, kink_vertex, right_vertex, radius
            )

            smoothing_arc_points = make_smoothing_arc(
                left_segment, left_offset, right_segment, right_offset
            )

            new_left_vertex, new_arc, new_right_vertex = make_smoothed_vertices(
                smoothing_arc_points, left_vertex, kink_vertex, right_vertex, left_segment, right_segment
            )

            vertices[i_kink_with_offset-1] = np.asarray(new_left_vertex)
            vertices[i_kink_with_offset] = np.asarray(new_right_vertex)
            vertices[i_kink_with_offset:i_kink_with_offset] = [np.asarray(new_arc)]

            index_offset += 1

        return ArcSpline(np.array(vertices), arc_spline.is_closed)

    # no kinks found, no smoothing must be done
    return arc_spline



from typing import Optional

from fibomat.layout.lattices.lattice_builder_base import LatticeBuilderBase
from fibomat.layout.lattices.lattice import Lattice
from fibomat.linalg import VectorLike, Vector


class LatticeBuilder(LatticeBuilderBase):
    _VectorClass = Vector

    def __init__(self, nu: int, nv: int, u: VectorLike, v: VectorLike, center: Optional[VectorLike] = None):
        super().__init__(nu, nv, u, v, center)

    @classmethod
    def generate_rect(cls, nu: int, nv: int, du: float, dv: float, center: Optional[VectorLike] = None):
        return cls(nu, nv, (float(du), 0), (0, -float(dv)), center)

    def to_lattice(self) -> Lattice:
        return Lattice(self._elements)




import difflib
from typing import Type, Dict

from fibomat.backend.backendbase import BackendBase


class Registry:
    """Class is used to register and store backends for exporting.

    Usually, you do not have to create an instance yourself and you rather
    should use the predefined instance :data:`registry`.
    """
    def __init__(self):
        self._backends: Dict[str, Type[BackendBase]] = {}

    @property
    def backends(self) -> Dict[str, Type[BackendBase]]:
        """Dict[str, Type[BackendBase]]: registered backends"""
        return self._backends

    def register(self, class_: Type[BackendBase], name: str) -> Type[BackendBase]:
        """Register the the backend class `class_` with name `name`.
        `name` has to be used in :meth:`get` to retrieve the backend
        class.

        .. note:: `name` can by arbitrary and any other registered
                  backend with the same name gets overwritten.

        Args:
            class_ (typing.Type[BackendBase]): backend class to be registered
            name (str): name of the backend

        Raises:
            TypeError: Raised if `class_` is not a subclass of
                       :class:`~fibomat.backend.backendbase.BackendBase`.
        """
        if not issubclass(class_, BackendBase):
            raise TypeError("class_ must be a subclass of BackendBase")

        self._backends[name] = class_

        return class_

    def get(self, name: str) -> Type[BackendBase]:
        """Returns the registered backend with name `name`

        Args:
            name (str): Name of backend

        Returns:
            typing.Type[BackendBase]: Backend class
        """

        if name in self._backends:
            return self._backends[name]
        else:
            matches = difflib.get_close_matches(name, self._backends.keys())
            error_massage = f'No backend found with name "{name}".'
            if matches:
                error_massage += f' Did you mean "{matches[0]}"?'
            raise KeyError(error_massage)


registry = Registry()
""":class:`Registry` instance"""

# def register_backend(name: str) -> typing.Callable:
#     """Use this function as a decorator to register a new backend for exporting::
#
#         @register_backend("mycoolexporter")
#         class MyCoolExporter(BaseBackend):
#             ...
#
#
#     Note that any other registered backend with the same name gets overwritten.
#
#     Args:
#         name (str): Name of the backend.
#     """
#     print('registering', name)
#     return functools.partial(registry.register, name=name)



from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.raster_styles.zero_d.prerasterized import PreRasterized
from fibomat.raster_styles import zero_d
from fibomat.raster_styles import one_d
from fibomat.raster_styles import two_d
from fibomat.raster_styles.scansequence import ScanSequence
#
# __all__ = ['RasterStyle', 'zero_d', 'one_d', 'PreRasterized', 'two_d', 'ScanSequence']



"""Provide a type annotation object for path like parameters.

Attributes:
    PathLike (str, or pathlib.Path): type annotation object for path like objects
"""
from typing import Union
import pathlib

PathLike = Union[str, pathlib.Path]
